import math
import statistics

import numpy as np
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "growth_factor",
    [1.0, 1.0000001, 1.01],
)
def test_calc_interval_probability_proxy(growth_factor):
    prior = hstrat.GeometricPrior(growth_factor)
    assert math.isclose(
        prior.CalcIntervalProbabilityProxy(0, 100),
        prior.CalcIntervalProbabilityProxy(0, 50)
        + prior.CalcIntervalProbabilityProxy(50, 100),
    )
    assert math.isclose(
        prior.CalcIntervalProbabilityProxy(89, 100),
        prior.CalcIntervalProbabilityProxy(89, 91)
        + prior.CalcIntervalProbabilityProxy(91, 100),
    )

    assert (
        pytest.approx(
            prior.CalcIntervalProbabilityProxy(99, 100)
            / prior.CalcIntervalProbabilityProxy(0, 1)
        )
        == growth_factor**99
    )

    assert (
        pytest.approx(
            prior.CalcIntervalProbabilityProxy(42, 43)
            / prior.CalcIntervalProbabilityProxy(1, 2)
        )
        == growth_factor**41
    )


@pytest.mark.parametrize(
    "growth_factor",
    [1.0, 1.0000001, 1.01],
)
def test_calc_interval_conditioned_mean(growth_factor):
    prior = hstrat.GeometricPrior(growth_factor)
    begin, end = 0, 100
    interval_width = 1

    intervals = [
        (interval_begin, interval_begin + interval_width)
        for interval_begin in range(begin, end, interval_width)
    ]

    samples = np.array(
        [
            statistics.mean((interval_begin, interval_end - 1))
            for interval_begin, interval_end in intervals
        ]
    )
    weights = np.array(
        [
            prior.CalcIntervalProbabilityProxy(interval_begin, interval_end)
            for interval_begin, interval_end in intervals
        ]
    )

    assert math.isclose(
        np.average(samples, weights=weights),
        prior.CalcIntervalConditionedMean(begin, end),
    )



import typing

import numpy as np

from ..._auxiliary_lib import (
    iter_monotonic_equivalencies_reverse,
    jit,
    reversed_range,
)
from ...frozen_instrumentation import HereditaryStratigraphicSpecimen
from .._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from .._impl_column._calc_rank_of_parity_segue_between import (
    calc_rank_of_parity_segue_between,
)
from ._iter_mutual_rank_indices import _compare_differentia_at_common_ranks

_reversed_range_jit = jit(nopython=True)(reversed_range)


@jit(nopython=True)
def _find_first_disparity_or_last_commonality(
    first_ranks: np.array,
    first_differentiae: np.array,
    second_ranks: np.array,
    second_differentiae: np.array,
) -> typing.Optional[typing.Tuple[int, int]]:
    for positions, comp in _compare_differentia_at_common_ranks(
        first_ranks, first_differentiae, second_ranks, second_differentiae
    ):
        if not comp:
            return positions
    return positions


@jit(nopython=True)
def _backtrack_n_equivalencies(
    first_ranks: np.array,
    second_ranks: np.array,
    start: typing.Tuple[int, int],
    n: int,
    strict: bool,
) -> typing.Tuple[int, int]:

    for i, positions in enumerate(
        iter_monotonic_equivalencies_reverse(
            first_ranks,
            second_ranks,
            start=start,
        )
    ):
        if i == n:
            return positions
    if strict:
        return None
    else:
        return (0, 0)


def calc_rank_of_parity_segue_between_naive(
    first: HereditaryStratigraphicSpecimen,
    second: HereditaryStratigraphicSpecimen,
    confidence_level_commonality: float,
    confidence_level_disparity: float,
) -> typing.Optional[int]:
    """Find first mismatching strata between columns.

    Implementation detail. Searches forward from index 0 for first mismatching
    retained differentiae between columns.
    """

    # terminology
    # rank: generation of deposition
    # loc: index among all strata
    # pos: index among common strata

    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )
    diff_bit_width = first.GetStratumDifferentiaBitWidth()

    if diff_bit_width > 64:
        # object-type numpy arrays breaks Numba nopython jit; send elsewhere
        return calc_rank_of_parity_segue_between(
            first,
            second,
            confidence_level_commonality=confidence_level_commonality,
            confidence_level_disparity=confidence_level_disparity,
        )

    first_vals = first.GetDifferentiaVals()
    second_vals = second.GetDifferentiaVals()
    first_ranks = first.GetRankIndex()
    second_ranks = second.GetRankIndex()

    first_pos, second_pos = _find_first_disparity_or_last_commonality(
        first_ranks, first_vals, second_ranks, second_vals
    )
    assert 0 <= first_pos < first.GetNumStrataRetained()
    assert 0 <= second_pos < second.GetNumStrataRetained()

    def calc_disparity(first_pos, second_pos):
        collision_plausibility_threshold = (
            calc_min_implausible_spurious_consecutive_differentia_collisions_between(
                first,
                second,
                significance_level=1.0 - confidence_level_disparity,
            )
            - 1
        )
        num_backtrack = collision_plausibility_threshold

        if first_vals[first_pos] == second_vals[second_pos]:
            if collision_plausibility_threshold == 0:
                if (
                    first.GetNumStrataDeposited()
                    == second.GetNumStrataDeposited()
                ):
                    return None
                else:
                    return min(
                        first.GetNumStrataDeposited(),
                        second.GetNumStrataDeposited(),
                    )
            else:
                num_backtrack -= 1

        if num_backtrack:
            assert first_ranks[first_pos] == second_ranks[second_pos]
            first_pos, second_pos = _backtrack_n_equivalencies(
                first_ranks,
                second_ranks,
                start=(first_pos, second_pos),
                n=num_backtrack,
                strict=False,
            )
            assert 0 <= first_pos < first.GetNumStrataRetained()
            assert 0 <= second_pos < second.GetNumStrataRetained()

        assert first_ranks[first_pos] == second_ranks[second_pos]
        # must convert to Python int; although integral,
        # numpy ints are experiencing unwanted conversion to floats
        assert isinstance(first_ranks[first_pos], np.integer)
        assert int(first_ranks[first_pos]) >= 0
        return int(first_ranks[first_pos])

    def calc_commonality(first_pos, second_pos):
        collision_plausibility_threshold = calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            first,
            second,
            significance_level=1.0 - confidence_level_commonality,
        )
        assert collision_plausibility_threshold > 0
        num_backtrack = collision_plausibility_threshold
        assert num_backtrack > 0

        if first_vals[first_pos] == second_vals[second_pos]:
            num_backtrack -= 1

        assert first_ranks[first_pos] == second_ranks[second_pos]
        # discount collision_implausibility_threshold - 1 common strata
        # as potential spurious differentia collisions
        res = _backtrack_n_equivalencies(
            first_ranks,
            second_ranks,
            start=(first_pos, second_pos),
            n=num_backtrack,
            strict=True,
        )
        if res is None:
            # not enough common strata to discount the possibility all
            # are spurious collisions with respect to the given confidence
            # level; conservatively conclude there is no common ancestor
            return res
        else:
            first_pos, second_pos = res
            assert 0 <= first_pos < first.GetNumStrataRetained()
            assert 0 <= second_pos < second.GetNumStrataRetained()
            assert first_ranks[first_pos] == second_ranks[second_pos]
            # must convert to Python int; although integral,
            # numpy ints are experiencing unwanted conversion to floats
            assert isinstance(first_ranks[first_pos], np.integer)
            assert int(first_ranks[first_pos]) >= 0
            return int(first_ranks[first_pos])

    return (
        calc_commonality(first_pos, second_pos)
        if confidence_level_commonality is not None
        else None,
        calc_disparity(first_pos, second_pos)
        if confidence_level_disparity is not None
        else None,
    )



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcRankAtColumnIndex:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcRankAtColumnIndex",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcRankAtColumnIndex", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcRankAtColumnIndex",
        policy: typing.Optional[PolicyCouplerBase],
        index: int,
        num_strata_deposited: typing.Optional[int],
    ) -> int:
        """After n strata have been deposited, what will the rank of the
        stratum at column index k be?

        Enables a HereditaryStratigraphicColumn using this predicate to
        optimize away storage of rank annotations on strata. Takes into the
        account the possibility for in-progress stratum depositions that haven't
        been reflected in num_strata_deposited.
        """
        return [
            0,
            num_strata_deposited - 1,
            num_strata_deposited,
        ][index]



"""Provides numba dict type or builtin bool type.

Type must be declared outside jit'ed function or numba fails.
"""

import typing

from ._is_in_coverage_run import is_in_coverage_run


class _shim:
    def empty(self: "_shim", *args, **kwargs) -> typing.Dict:
        return dict()


try:
    from numba.typed import Dict
except (ImportError, ModuleNotFoundError):
    jit_numba_dict_t = _shim()
else:
    if is_in_coverage_run():
        jit_numba_dict_t = _shim()
    else:  # pragma: no cover
        jit_numba_dict_t = Dict



import itertools as it
import unittest

import numpy as np

from hstrat._auxiliary_lib import all_unique


class TestAllUnique(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test(self):

        for num_unique, num_duplications, shuffle, in it.product(
            np.arange(10),
            np.arange(1, 10),
            [True, False],
        ):
            test_arr = np.array(
                [
                    val
                    for val in range(num_unique)
                    for __ in range(num_duplications)
                ]
            )

            if shuffle:
                np.random.shuffle(test_arr)

            assert (num_unique == len(test_arr)) == all_unique(test_arr)


if __name__ == "__main__":
    unittest.main()



import os

import alifedata_phyloinformatics_convert as apc
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import alifestd_validate

from . import _impl as impl

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    [
        "arbitrary",
        "uniform",
        hstrat.ExponentialPrior(1.01),
    ],
)
def test_determinism(orig_tree, retention_policy, wrap, estimator, prior):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    first_reconst = hstrat.build_tree_nj(extant_population, estimator, prior)
    for _rep in range(3):
        _ = _rep
        second_reconst = hstrat.build_tree_nj(
            [wrap(col) for col in extant_population],
            estimator,
            prior,
        )
        assert first_reconst.equals(second_reconst)


@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(4),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_reconstructed_taxon_labels(orig_tree, retention_policy, wrap):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )
    taxon_labels = [str(id(x)) for x in extant_population]

    reconst_df = hstrat.build_tree_nj(
        [*map(wrap, extant_population)],
        "maximum_likelihood",
        "arbitrary",
        taxon_labels=taxon_labels,
        negative_origin_time_correction_method="shift",
    )
    assert "taxon_label" in reconst_df
    assert set(taxon_labels) < set(reconst_df["taxon_label"])

    reconst_df = hstrat.build_tree_nj(
        [*map(wrap, extant_population)],
        "maximum_likelihood",
        "arbitrary",
        negative_origin_time_correction_method="shift",
    )
    assert "taxon_label" in reconst_df
    assert len(reconst_df["taxon_label"].unique()) == len(
        reconst_df["taxon_label"]
    )
    assert set(map(str, range(len(extant_population)))) < set(
        reconst_df["taxon_label"]
    )


@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        # nj flunks building this tree with quartet distance 1.0
        # over all retention policies
        # upgma flunks too (but build_tree_trie doesn't)
        # impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(2),
        hstrat.recency_proportional_resolution_algo.Policy(2),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_reconstructed_dist(orig_tree, retention_policy, wrap):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    reconst_df = hstrat.build_tree_nj(
        [wrap(col) for col in extant_population],
        "maximum_likelihood",
        "arbitrary",
        taxon_labels=(leaf.taxon.label for leaf in orig_tree.leaf_node_iter()),
    )
    assert "origin_time" in reconst_df

    assert alifestd_validate(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )

    assert len(list(reconst_tree.leaf_node_iter())) == len(extant_population)
    sorted_leaf_nodes = sorted(
        reconst_tree.leaf_node_iter(), key=lambda x: int(x.taxon.label)
    )
    assert {
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    } == {
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    }
    assert sorted(
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    ) == sorted(
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    )
    assert [
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    ] == [
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    ]

    reconstruction_error = impl.tree_quartet_distance(
        orig_tree,
        reconst_tree,
    )
    # arbitrary threshold...  quartet dist expectations: random 0.75, worst 1.0
    assert 0.0 <= reconstruction_error < 0.4



import typing

from ...._detail import PolicyCouplerBase
from ..._PolicySpec import PolicySpec
from ..._impl import get_retained_ranks


class GenDropRanks:
    """Functor to implement the approximate space-filling MRCA-recency-
    proportional resolution stratum retention policy, for use with
    HereditaryStratigraphicColumn.

    This functor enacts the approximate space-filling MRCA-recency-
    proportional resolution stratum retention policy by specifying the set of
    strata ranks that should be purged from a hereditary stratigraphic column
    when the nth stratum is deposited.
    """

    def __init__(
        self: "GenDropRanks",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "GenDropRanks", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "GenDropRanks",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        retained_ranks: typing.Optional[typing.Iterable[int]],
    ) -> typing.Iterator[int]:
        """Decide which strata within the stratagraphic column should be purged.

        Every time a new stratum is deposited, this method is called to
        determine which strata should be purged. All strata at ranks yielded
        from this functor are immediately purged from the column, meaning that
        for a stratum to persist it must not be yielded by this functor each
        and every time a new stratum is deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.
        retained_ranks : iterator over int
            An iterator over ranks of strata currently retained within the
            hereditary stratigraphic column. Not used in this functor.

        Returns
        -------
        iterator over int
                The ranks of strata that should be purged from the hereditary
            stratigraphic column at this deposition step.

        See Also
        --------
        depth_proportional_resolution_tapered_algo:
            For details on the rationale, implementation, and guarantees of the
            tapered depth-proportional resolution stratum retention policy.
        """
        spec = policy.GetSpec()

        prev_retained_ranks = get_retained_ranks(
            spec.GetDegree(),
            spec.GetInterspersal(),
            num_stratum_depositions_completed,
        )

        cur_retained_ranks = get_retained_ranks(
            spec.GetDegree(),
            spec.GetInterspersal(),
            num_stratum_depositions_completed + 1,
        )

        # take set difference between prev retained ranks and cur retained ranks
        yield from (prev_retained_ranks - cur_retained_ranks)



def num_to_condemn_reference(
    recency_proportional_resolution: int,
    num_stratum_depositions_completed: int,
) -> int:
    """How many strata should be eliminated after
    num_stratum_depositions_completed have been deposited and the
    num_stratum_depositions_completed + 1'th deposition is in progress?

    Used to implement functor's __call__ method specifying which ranks should
    be purged during this stratum deposition. This expression for exact number
    deposited was extrapolated from
        * resolution = 0, <https://oeis.org/A001511>
        * resolution = 1, <https://oeis.org/A091090>
    and is unit tested extensively.
    """
    resolution = recency_proportional_resolution

    if num_stratum_depositions_completed % 2 == 1:
        return 0
    elif num_stratum_depositions_completed < 2 * (resolution + 1):
        return 0
    else:
        return 1 + num_to_condemn_reference(
            resolution,
            num_stratum_depositions_completed // 2,
        )



import numpy as np

from hstrat._auxiliary_lib import iter_monotonic_equivalencies


def test_empty_arrays():
    first = np.array([])
    second = np.array([])
    res = [*iter_monotonic_equivalencies(first, second)]
    assert res == []
    for i, j in res:
        assert first[i] == second[j]


def test_one_element_arrays():
    first = np.array([1])
    second = np.array([1])
    res = [*iter_monotonic_equivalencies(first, second)]
    assert res == [(0, 0)]
    for i, j in res:
        assert first[i] == second[j]


def test_identical_arrays():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([1, 2, 3, 4, 5])
    res = [*iter_monotonic_equivalencies(first, second)]
    assert res == [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)]
    for i, j in res:
        assert first[i] == second[j]


def test_non_overlapping_arrays():
    first = np.array([1, 3, 5, 7, 9])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies(first, second)]
    assert res == []
    for i, j in res:
        assert first[i] == second[j]


def test_some_overlapping_arrays():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies(first, second)]
    assert res == [(1, 0), (3, 1)]
    for i, j in res:
        assert first[i] == second[j]


def test_non_decreasing_arrays():
    first = np.array([1, 2, 3, 3, 3, 4, 5, 7])
    second = np.array([2, 3, 3, 4, 4, 4, 6, 7])
    res = [*iter_monotonic_equivalencies(first, second)]
    assert res == [(1, 0), (2, 1), (3, 2), (5, 3), (7, 7)]
    for i, j in res:
        assert first[i] == second[j]


def test_start_parameter1():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies(first, second, start=(1, 0))]
    assert res == [(1, 0), (3, 1)]
    for i, j in res:
        assert first[i] == second[j]


def test_start_parameter2():
    first = np.array([1, 2, 3, 4, 5])
    second = np.array([2, 4, 6, 8, 10])
    res = [*iter_monotonic_equivalencies(first, second, start=(2, 0))]
    assert res == [(3, 1)]
    for i, j in res:
        assert first[i] == second[j]



import json
import logging

import pytest

from hstrat import genome_instrumentation, hstrat
from hstrat._auxiliary_lib import log_once_in_a_row


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_records(
    impl,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    records = hstrat.col_to_records(column)
    assert records == hstrat.col_to_records(column)
    for entry in [
        "policy_algo",
        "policy",
        "num_strata_deposited",
        "differentiae",
        "differentia_bit_width",
        "hstrat_version",
    ]:
        assert entry in records
        assert isinstance(records[entry], (int, str))

    assert "policy_spec" in records
    assert isinstance(records["policy_spec"], dict)


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_records_then_from_records(
    impl,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    assert hstrat.col_to_records(column) == hstrat.col_to_records(column)
    reconstituted = hstrat.col_from_records(hstrat.col_to_records(column))
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
    ):
        assert reconstituted == column
    else:
        assert hstrat.col_to_records(reconstituted) == hstrat.col_to_records(
            column
        )


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_records_then_from_records_with_annotations(
    impl,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        initial_stratum_annotation=9,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for gen in range(num_deposits):
        column.DepositStratum(annotation=gen)

    assert hstrat.col_to_records(column) == hstrat.col_to_records(column)
    reconstituted = hstrat.col_from_records(hstrat.col_to_records(column))
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
    ):
        assert reconstituted == column
    else:
        assert hstrat.col_to_records(reconstituted) == hstrat.col_to_records(
            column
        )


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_records_then_from_records_json(
    impl,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    records = hstrat.col_to_records(column)
    json_str = json.dumps(records)
    reconstituted = hstrat.col_from_records(json.loads(json_str))
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
    ):
        assert reconstituted == column
    else:
        assert hstrat.col_to_records(reconstituted) == hstrat.col_to_records(
            column
        )


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_records_version_warning(
    impl,
    retention_policy,
    ordered_store,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    # adapted from https://stackoverflow.com/a/48113200
    with caplog.at_level(logging.INFO):
        # clear caplog
        caplog.clear()

        # create a record with a phony version
        record = hstrat.col_to_records(column)
        record["hstrat_version"] = "0.0.0"

        # reconstruct phony record
        hstrat.col_from_records(record)

        # it should have logged a warning
        assert len(caplog.records)

        # log something to keep the lru cache from supressing the above calls
        log_once_in_a_row("foo")



import numpy as np

from ._jit import jit


@jit(nopython=True)
def apply_swaps(
    arr: np.array, swapfrom_idxs: np.array, swapto_idxs: np.array
) -> None:
    """Swap elements in `arr` at `swapfrom_idxs` indices with elements at
    `swapto_idxs` indices.

    The elements at the first index in `swapfrom_idxs` and the first index in
    `swapfrom_idxs` are swapped, then elements at the second indices in the
    the argument arrays, and so on. Swaps are applied to `arr` inplace.
    """
    for a, b in zip(swapfrom_idxs, swapto_idxs):
        arr[a], arr[b] = arr[b], arr[a]



import itertools as it
import typing
import warnings

import pandas as pd
import sortedcontainers as sc

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_has_multiple_roots import alifestd_has_multiple_roots
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_mark_leaves import alifestd_mark_leaves
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_ot_mrca_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
) -> pd.DataFrame:
    """Add columns `ot_mrca_id`, `ot_mrca_time_of`, and `ot_mrca_time_since`,
    giving information about mrca of extant organisms at organism `origin_time`.

    A chronological sort will be applied if `phylogeny_df` is not
    chronologically sorted. Reindexing may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    # setup
    if alifestd_has_multiple_roots(phylogeny_df):
        raise NotImplementedError()

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)
    phylogeny_df = alifestd_mark_leaves(phylogeny_df, mutate=True)

    contig = alifestd_has_contiguous_ids(phylogeny_df)
    if contig:
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        warnings.warn("mark_ot_mrca_asexual may be slow with uncontiguous ids")
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["ot_mrca_id"] = phylogeny_df["id"].max() + 1
    phylogeny_df["ot_mrca_time_of"] = phylogeny_df["origin_time"].max()
    phylogeny_df["ot_mrca_time_since"] = phylogeny_df["origin_time"].max()

    df = phylogeny_df
    df["bwd_origin_time"] = -df["origin_time"]

    # do calculation
    running_mrca_id = max(
        df["id"],
        default=None,
        key=lambda i: (df.at[i, "origin_time"], df.index.get_loc(i)),
    )  # initial value

    # for optimization
    ot_mrca_ids = []
    ot_mrca_times_of = []
    ot_mrca_times_since = []
    indices = []
    key = df.index.get_loc if not contig else None
    for bwd_origin_time, group in progress_wrap(df.groupby("bwd_origin_time")):
        earliest_id = (
            group["id"].min() if contig else min(group["id"], key=key)
        )

        leaf_mask = group["is_leaf"]
        lineages = sc.SortedSet(
            (*group.loc[leaf_mask, "id"], earliest_id, running_mrca_id),
            key=key,
        )
        while len(lineages) > 1:
            oldest = lineages.pop(-1)
            replacement = df.at[oldest, "ancestor_id"]
            assert replacement != oldest
            lineages.add(replacement)

        (mrca_id,) = lineages
        running_mrca_id = mrca_id

        # set column values
        mrca_time = df.at[mrca_id, "origin_time"]

        size = len(group)
        indices.extend(group["id"])
        ot_mrca_ids.extend(it.repeat(mrca_id, size))
        ot_mrca_times_of.extend(it.repeat(mrca_time, size))
        ot_mrca_times_since.extend(
            it.repeat(-bwd_origin_time - mrca_time, size),
        )

    df.loc[indices, "ot_mrca_id"] = ot_mrca_ids
    df.loc[indices, "ot_mrca_time_of"] = ot_mrca_times_of
    df.loc[indices, "ot_mrca_time_since"] = ot_mrca_times_since
    df.drop("bwd_origin_time", axis=1, inplace=True)
    return df



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_last_retained_commonality_between import (
    calc_rank_of_last_retained_commonality_between,
)


def calc_ranks_since_last_retained_commonality_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """Determine generations since MRCA with particular confidence.

    How many depositions have elapsed along the focal column's line of
    descent since the last matching strata at the same rank between focal and
    other?

    Returns None if no common ancestor is shared between focal and other.

    Parameters
    ----------
    confidence_level : float, optional
        With what probability should the true number of ranks since the
        last commonality with other be less than the calculated estimate?
        Default 0.95.

    Notes
    -----
    If confidence_level < 0.5, then the true number of ranks since the last
    commonality with other is guaranteed greater than or equal to the
    calculated estimate.
    """
    assert 0.0 <= confidence_level <= 1.0

    last_common_rank = calc_rank_of_last_retained_commonality_between(
        focal,
        other,
        confidence_level=confidence_level,
    )
    if last_common_rank is None:
        return None
    else:
        assert focal.GetNumStrataDeposited()
        res = focal.GetNumStrataDeposited() - 1 - last_common_rank
        assert 0 <= res < focal.GetNumStrataDeposited()
        return res



import typing

# use string alias due to circular module initialization issue
HereditaryStratigraphicArtifact = typing.Union[
    "HereditaryStratigraphicColumn",
    "HereditaryStratigraphicSpecimen",
]



import anytree
import pandas as pd

from hstrat._auxiliary_lib import anytree_tree_to_alife_dataframe


def test_singleton():
    original_tree = anytree.Node("beep")
    assert original_tree.name == "beep"
    converted_df = anytree_tree_to_alife_dataframe(original_tree)

    expected_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[None]"],
            "name": ["beep"],
        }
    )

    assert converted_df.equals(expected_df)


def test_simple_tree():
    udo = anytree.Node("Udo", id=0)
    marc = anytree.Node("Marc", parent=udo, id=1111)
    anytree.Node("Lian", parent=marc, id=2222)
    dan = anytree.Node("Dan", parent=udo, id=333)
    anytree.Node("Jet", parent=dan, id=444)
    anytree.Node("Jan", parent=dan, id=555)
    anytree.Node("Joe", parent=dan, id=666)

    converted_df = anytree_tree_to_alife_dataframe(udo)
    expected_df = pd.DataFrame(
        {
            "id": [0, 1111, 333, 2222, 444, 555, 666],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1111]",
                "[333]",
                "[333]",
                "[333]",
            ],
            "name": ["Udo", "Marc", "Dan", "Lian", "Jet", "Jan", "Joe"],
        }
    )

    assert converted_df.equals(expected_df)


def test_simple_tree_origin_times():
    udo = anytree.Node("Udo", id=0, edge_length=1)
    marc = anytree.Node("Marc", parent=udo, id=1111, edge_length=1)
    anytree.Node("Lian", parent=marc, id=2222, edge_length=1)
    dan = anytree.Node("Dan", parent=udo, id=333, edge_length=2)
    anytree.Node("Jet", parent=dan, id=444, edge_length=4)
    anytree.Node("Jan", parent=dan, id=555, edge_length=5)
    anytree.Node("Joe", parent=dan, id=666, edge_length=6)

    converted_df = anytree_tree_to_alife_dataframe(udo)
    expected_df = pd.DataFrame(
        {
            "id": [0, 1111, 333, 2222, 444, 555, 666],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1111]",
                "[333]",
                "[333]",
                "[333]",
            ],
            "edge_length": [1, 1, 2, 1, 4, 5, 6],
            "name": ["Udo", "Marc", "Dan", "Lian", "Jet", "Jan", "Joe"],
            "origin_time": [1, 2, 3, 3, 7, 8, 9],
        }
    )

    assert converted_df.equals(expected_df)



import numpy as np

from ._jit import jit


@jit(nopython=True)
def unfurl_lineage_with_contiguous_ids(
    ancestor_ids: np.array, leaf_id: int
) -> np.array:
    """List leaf id and its ancestor id sequence through tree root.

    Assumes that each organism's ancestor id is located at the index position
    in `ancestor_ids` corresponding to its own id.
    """
    id_ = leaf_id
    res = list()
    while True:
        res.append(id_)
        next_id = ancestor_ids[id_]
        if id_ == next_id:
            break
        id_ = next_id

    return np.array(res)



import pickle
import tempfile

import pytest

from hstrat.hstrat import depth_proportional_resolution_tapered_algo


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(depth_proportional_resolution):
    spec = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution
    )
    assert spec == spec
    assert spec == depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution
    )
    assert not spec == depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution + 1
    )


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_GetEvalCtor(depth_proportional_resolution):
    # hstrat. is needed for eval()
    from hstrat import hstrat  # noqa

    spec = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution
    )
    eval_ctor = spec.GetEvalCtor()
    assert eval_ctor.startswith(
        "hstrat.depth_proportional_resolution_tapered_algo.PolicySpec("
    )
    assert eval_ctor.endswith(")")
    reconstituted = eval(eval_ctor)  # noqa
    assert spec == reconstituted


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_pickle(depth_proportional_resolution):
    original = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution
    )
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.parametrize(
    "depth_proportional_resolution",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_GetFixedResolution(depth_proportional_resolution):
    spec = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution
    )
    assert (
        spec.GetDepthProportionalResolution() == depth_proportional_resolution
    )


def test_GetAlgoIdentifier():
    depth_proportional_resolution = 1
    spec = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution,
    )
    assert spec.GetAlgoIdentifier()


def test_GetAlgoTitle():
    depth_proportional_resolution = 1
    spec = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution,
    )
    assert spec.GetAlgoTitle()


def test_repr():
    depth_proportional_resolution = 1
    spec = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution,
    )
    assert str(depth_proportional_resolution) in repr(spec)
    assert spec.GetAlgoIdentifier() in repr(spec)


def test_str():
    depth_proportional_resolution = 1
    spec = depth_proportional_resolution_tapered_algo.PolicySpec(
        depth_proportional_resolution,
    )
    assert str(depth_proportional_resolution) in str(spec)
    assert spec.GetAlgoTitle() in str(spec)



"""Convenient library repack as a flat namespace."""

from .._auxiliary_lib import get_hstrat_version
from ..frozen_instrumentation import *  # noqa: F401
from ..genome_instrumentation import *  # noqa: F401
from ..juxtaposition import *  # noqa: F401
from ..phylogenetic_inference import *  # noqa: F401
from ..serialization import *  # noqa: F401
from ..stratum_retention_strategy import *  # noqa: F401
from ..stratum_retention_viz import *  # noqa: F401
from ..test_drive import *  # noqa: F401

__version__ = get_hstrat_version()



import math

import numpy as np
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "growth_factor",
    [
        1.0,
        1.0000001,
        1.000001,
        1.0001,
        1.01,
        1.5,
    ],
)
def test_calc_interval_probability_proxy(growth_factor):
    prior = hstrat.ExponentialPrior(growth_factor)
    assert math.isclose(
        prior.CalcIntervalProbabilityProxy(0, 100),
        prior.CalcIntervalProbabilityProxy(0, 50)
        + prior.CalcIntervalProbabilityProxy(50, 100),
    )
    assert math.isclose(
        prior.CalcIntervalProbabilityProxy(89, 100),
        prior.CalcIntervalProbabilityProxy(89, 91)
        + prior.CalcIntervalProbabilityProxy(91, 100),
    )

    assert (
        pytest.approx(
            prior.CalcIntervalProbabilityProxy(99, 100)
            / prior.CalcIntervalProbabilityProxy(0, 1)
        )
        == growth_factor**99
    )

    assert (
        pytest.approx(
            prior.CalcIntervalProbabilityProxy(42, 43)
            / prior.CalcIntervalProbabilityProxy(1, 2)
        )
        == growth_factor**41
    )
    assert prior.CalcIntervalProbabilityProxy(0, 1) >= 0


@pytest.mark.parametrize(
    "interval_width",
    [1, 2],
)
@pytest.mark.parametrize(
    "growth_factor",
    [1.0, 1.0000001, 1.01, 1.05],
)
def test_calc_interval_conditioned_mean(interval_width, growth_factor):
    prior = hstrat.ExponentialPrior(growth_factor)
    begin, end = 0, 10000

    intervals = [
        (interval_begin, interval_begin + interval_width)
        for interval_begin in range(begin, end, interval_width)
    ]

    samples = np.array(
        [
            # interval_begin
            prior.CalcIntervalConditionedMean(interval_begin, interval_end)
            for interval_begin, interval_end in intervals
        ]
    )
    weights = np.array(
        [
            prior.CalcIntervalProbabilityProxy(interval_begin, interval_end)
            for interval_begin, interval_end in intervals
        ]
    )

    # some discretization bias is expected
    assert np.average(samples, weights=weights) == pytest.approx(
        prior.CalcIntervalConditionedMean(begin, end), abs=0.5
    )
    assert 0 <= prior.CalcIntervalConditionedMean(0, 1) <= 1



import typing

from ...._detail import PolicyCouplerBase
from ...._impl import GenDropRanksFromPredKeepRank
from ....geom_seq_nth_root_algo._enact._GenDropRanks_._FromPredKeepRank import (
    _PredKeepRank as gsnra_PredKeepRank,
)
from ....recency_proportional_resolution_algo._enact._GenDropRanks_._FromPredKeepRank import (
    _PredKeepRank as rpra_PredKeepRank,
)
from ..._PolicySpec import PolicySpec
from ..._impl import pick_policy


class _PredKeepRank:
    """Functor to implement the recency-proportional resolution stratum
    retention policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the recency-proportional resolution policy by
    specifying the set of strata ranks that should be purged from a hereditary
    stratigraphic column when the nth stratum is deposited.
    """

    def __init__(
        self: "_PredKeepRank",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "_PredKeepRank", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "_PredKeepRank",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        stratum_rank: int,
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.

        See Also
        --------
        recency_proportional_resolution_algo:
            For details on the rationale, implementation, and guarantees of the
            recency-proportional resolution stratum retention policy.
        """
        dispatched_policy = pick_policy(
            policy.GetSpec().GetSizeCurb(),
            num_stratum_depositions_completed + 1,
        )
        return {
            "recency_proportional_resolution_algo": rpra_PredKeepRank,
            "geom_seq_nth_root_algo": gsnra_PredKeepRank,
        }[dispatched_policy.GetSpec().GetAlgoIdentifier()]._do_call(
            dispatched_policy, num_stratum_depositions_completed, stratum_rank
        )


FromPredKeepRank = GenDropRanksFromPredKeepRank(_PredKeepRank)



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_is_chronologically_ordered,
    alifestd_make_empty,
    alifestd_mark_ot_mrca_asexual,
    alifestd_to_working_format,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        ),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    assert alifestd_is_chronologically_ordered(phylogeny_df)
    original = phylogeny_df.copy()

    result = alifestd_mark_ot_mrca_asexual(phylogeny_df)

    assert alifestd_validate(result)
    assert alifestd_is_chronologically_ordered(result)
    assert original.equals(phylogeny_df)

    assert all(original["id"].min() <= result["ot_mrca_id"])
    assert all(result["ot_mrca_id"] <= original["id"].max())

    assert all(original["origin_time"].min() <= result["ot_mrca_time_of"])
    assert all(result["ot_mrca_time_of"] <= original["origin_time"].max())

    assert all(0 <= result["ot_mrca_time_since"])
    assert all(result["ot_mrca_time_since"] <= original["origin_time"].max())


def test_empty():
    mt = alifestd_make_empty()
    mt["origin_time"] = None
    res = alifestd_mark_ot_mrca_asexual(mt)
    assert "ot_mrca_id" in res
    assert "ot_mrca_time_of" in res
    assert "ot_mrca_time_since" in res
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "origin_time": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_ot_mrca_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    assert result_df.loc[0, "ot_mrca_id"] == 0
    assert result_df.loc[1, "ot_mrca_id"] == 1
    assert result_df.loc[2, "ot_mrca_id"] == 2

    assert result_df.loc[0, "ot_mrca_time_of"] == 0
    assert result_df.loc[1, "ot_mrca_time_of"] == 1
    assert result_df.loc[2, "ot_mrca_time_of"] == 2

    assert result_df.loc[0, "ot_mrca_time_since"] == 0
    assert result_df.loc[1, "ot_mrca_time_since"] == 0
    assert result_df.loc[2, "ot_mrca_time_since"] == 0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 3, 2],
            "origin_time": [1, 0, 2, 3],
            "ancestor_list": ["[0]", "[None]", "[0]", "[3]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_ot_mrca_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[0, "ot_mrca_id"] == 0
    assert result_df.loc[1, "ot_mrca_id"] == 0
    assert result_df.loc[3, "ot_mrca_id"] == 3
    assert result_df.loc[2, "ot_mrca_id"] == 2

    assert result_df.loc[0, "ot_mrca_time_of"] == 0
    assert result_df.loc[1, "ot_mrca_time_of"] == 0
    assert result_df.loc[3, "ot_mrca_time_of"] == 2
    assert result_df.loc[2, "ot_mrca_time_of"] == 3

    assert result_df.loc[0, "ot_mrca_time_since"] == 0
    assert result_df.loc[1, "ot_mrca_time_since"] == 1
    assert result_df.loc[3, "ot_mrca_time_since"] == 0
    assert result_df.loc[2, "ot_mrca_time_since"] == 0

    if not mutate:
        assert original_df.equals(phylogeny_df)



import logging

import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import (
    alifestd_is_chronologically_ordered,
    alifestd_splay_polytomies,
    alifestd_validate,
    seed_random,
)


@pytest.mark.parametrize("tree_seed", [1, 2, 3])
@pytest.mark.parametrize("tournament_size", [1, 4])
@pytest.mark.parametrize("differentia_width", [1, 2, 8, 64])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(10),
        # hstrat.depth_proportional_resolution_tapered_algo.Policy(10),
    ],
)
def test_reconstructed_mrca_fuzz(
    tree_seed, tournament_size, differentia_width, retention_policy
):

    seed_random(tree_seed)
    tree = hstrat.evolve_fitness_trait_population(
        population_size=128,
        iter_epochs=False,
        num_islands=1,
        num_niches=1,
        num_generations=256,
        tournament_size=tournament_size,
        p_island_migration=0,
        p_niche_invasion=0,
        share_common_ancestor=True,
    )

    test_population = hstrat.descend_template_phylogeny_alifestd(
        tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_width,
        ),
    )
    control_population = hstrat.descend_template_phylogeny_alifestd(
        tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
            stratum_differentia_bit_width=64,
        ),
    )

    control_df = alifestd_splay_polytomies(
        hstrat.build_tree_trie(control_population)
    )
    assert alifestd_validate(control_df)
    assert alifestd_is_chronologically_ordered(control_df)

    out_dfs = map(
        alifestd_splay_polytomies,
        hstrat.build_tree_trie_ensemble(
            test_population,
            trie_postprocessors=[
                hstrat.AssignOriginTimeNodeRankTriePostprocessor(),
                hstrat.AssignOriginTimeNaiveTriePostprocessor(),
                hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
                    prior=hstrat.ArbitraryPrior(),
                ),
                hstrat.CompoundTriePostprocessor(
                    postprocessors=[
                        hstrat.PeelBackConjoinedLeavesTriePostprocessor(),
                        hstrat.AssignOriginTimeExpectedValueTriePostprocessor(
                            prior=hstrat.ArbitraryPrior()
                        ),
                    ],
                ),
                hstrat.CompoundTriePostprocessor(
                    postprocessors=[
                        hstrat.SampleAncestralRollbacksTriePostprocessor(
                            seed=tree_seed
                        ),
                        hstrat.AssignOriginTimeNaiveTriePostprocessor(),
                    ],
                ),
            ],
        ),
    )

    test_dfs = dict(
        zip(
            ["rank", "naive", "expected", "peeled_expected", "rollback"],
            out_dfs,
        )
    )

    true_mean_origin_time = control_df["origin_time"].mean()
    test_errs = dict()
    for postprocessor, test_df in test_dfs.items():
        assert alifestd_validate(test_df)
        assert alifestd_is_chronologically_ordered(test_df)
        test_mean_origin_time = test_df["origin_time"].mean()
        test_err = test_mean_origin_time - true_mean_origin_time
        test_errs[postprocessor] = test_err
        logging.info(f"{postprocessor} err {test_err}")

    # don't put too much stock into this test...
    # expected/naive should be better than rank most of the time in these tests
    # but not necessarily all the time
    assert abs(test_errs["peeled_expected"]) <= abs(test_errs["expected"])
    assert abs(test_errs["peeled_expected"]) <= abs(test_errs["rank"]) + 0.02
    if differentia_width == 64:
        assert abs(test_errs["naive"]) < abs(test_errs["rank"]) + 0.02



import unittest

import numpy as np
import pandas as pd

from hstrat._auxiliary_lib import swap_rows_and_indices


class TestOmitLast(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_default_index(self):

        df = pd.DataFrame(
            np.array([[1, 2], [3, 4], [5, 6]]),
            columns=["a", "b"],
        )

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, 0, 0)
        assert df_.equals(df)

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, 0, 1)
        target = pd.DataFrame(
            np.array([[3, 4], [1, 2], [5, 6]]),
            columns=["a", "b"],
            index=[1, 0, 2],
        )

        df_ = swap_rows_and_indices(df_, 1, 0)
        assert df_.equals(df)

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, 0, 1)
        df_ = swap_rows_and_indices(df_, 1, 2)
        target = pd.DataFrame(
            np.array([[5, 6], [1, 2], [3, 4]]),
            columns=["a", "b"],
            index=[2, 0, 1],
        )

        assert df_.equals(target)

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, 0, 2)
        target = pd.DataFrame(
            np.array([[5, 6], [3, 4], [1, 2]]),
            columns=["a", "b"],
            index=[2, 1, 0],
        )
        assert df_.equals(target)

    def test_custom_index(self):

        df = pd.DataFrame(
            np.array([[1, 2], [3, 4], [5, 6]]),
            columns=["a", "b"],
            index=["A", "B", "C"],
        )

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, "A", "A")
        assert df_.equals(df)

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, "A", "B")
        target = pd.DataFrame(
            np.array([[3, 4], [1, 2], [5, 6]]),
            columns=["a", "b"],
            index=["B", "A", "C"],
        )
        assert df_.equals(target)

        df_ = swap_rows_and_indices(df_, "B", "A")
        assert df_.equals(df)

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, "A", "B")
        df_ = swap_rows_and_indices(df_, "B", "C")
        target = pd.DataFrame(
            np.array([[5, 6], [1, 2], [3, 4]]),
            columns=["a", "b"],
            index=["C", "A", "B"],
        )
        assert df_.equals(target)

        df_ = df.copy()
        df_ = swap_rows_and_indices(df_, "A", "C")
        target = pd.DataFrame(
            np.array([[5, 6], [3, 4], [1, 2]]),
            columns=["a", "b"],
            index=["C", "B", "A"],
        )
        assert df_.equals(target)


if __name__ == "__main__":
    unittest.main()



"""Implementation helpers."""

from ._descend_unifurcations import descend_unifurcations
from ._load_dendropy_tree import load_dendropy_tree
from ._setup_dendropy_tree import setup_dendropy_tree
from ._tree_quartet_distance import tree_quartet_distance
from ._tree_unweighted_robinson_foulds_distance import (
    tree_unweighted_robinson_foulds_distance,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "descend_unifurcations",
    "load_dendropy_tree",
    "setup_dendropy_tree",
    "tree_quartet_distance",
    "tree_unweighted_robinson_foulds_distance",
]



import typing

from keyname import keyname as kn
from matplotlib import animation as mpl_animation
from matplotlib import pyplot as plt
from slugify import slugify
from tqdm import tqdm

from ..plot._policy_panel_plot import policy_panel_plot


def policy_panel_animate(
    stratum_retention_policy: typing.Any,
    num_generations: int,
    do_show: bool = False,
    save_as: typing.Optional[str] = None,
) -> mpl_animation.FuncAnimation:
    """Animate evolution of column and its properties under a retention policy.

    Parameters
    ----------
    stratum_retention_policy : any
        Object specifying stratum retention policy.
    num_generations : int
        Number of generations to plot.
    do_show : bool, optional
        Whether to show() the plot automatically.
    save_as : str, optional
        If set, save animation as file type specified.
    """
    fig = plt.figure(figsize=(8, 6), dpi=80)

    def update_func(frame) -> None:
        for ax in fig.axes:
            fig.delaxes(ax)
        fig.clear()
        policy_panel_plot(
            stratum_retention_policy=stratum_retention_policy,
            num_generations=frame,
            do_show=False,
            fig=fig,
        )

    res = mpl_animation.FuncAnimation(
        fig,
        update_func,
        frames=range(num_generations),
        interval=500,
        blit=False,
    )

    if save_as is not None:
        fname = kn.pack(
            {
                "a": "policy_panel_plot",
                "num_generations": num_generations,
                "policy": slugify(str(stratum_retention_policy)),
                "ext": f'.{save_as.strip(".")}',
            }
        )
        print(f"saving animation to {fname}")
        progress = tqdm(total=num_generations)
        res.save(
            fname,
            fps=5,
            writer="imagemagick",
            progress_callback=lambda *args, **kwargs: progress.update(),
        )

    if do_show:
        plt.show()

    return res



import collections
import types
import typing

import numpy as np
import pandas as pd

from ..._auxiliary_lib import apply_swaps, count_unique
from ._GarbageCollectingPhyloTracker_ import _discern_referenced_rows


class GarbageCollectingPhyloTracker:
    """Data structure to enable perfect tracking over a fixed-size population
    with synchronous generations.

    Designed to provide low-overhead tracking. Instead of representing organism
    records as independent objects (which each require an independent
    allocations and, on lineage extinction, deletions), stores organism records
    as rows within a numpy array. Partial garbage collection at regular intervals compacts recent record entries to discard extinct lineages.

    Includes organism population loc and trait values in phylogenetic record.
    """

    _population_size: int
    _working_buffer_size: int

    _parentage_buffer: np.array  # [int]
    """1d array with entries representing individual organisms and stored
    value at each entry corresponding to index of entry representing that
    organism's parent."""

    _num_records: int
    """How many entries are stored in `_parentage_buffer`? May be less than
    `len(_parentage_buffer)`."""

    # parentage_buffer:
    #
    #                 +----------------+
    #  *gen 0*        | parent_row_idx |
    #                 | parent_row_idx |
    #
    #                 ...
    #  *gen 1*        | parent_row_idx |
    #                 | parent_row_idx |
    #
    #                 ...
    #                 | parent_row_idx |
    # num_records --> | UNINITIALIZED  |
    #                 | UNINITIALIZED  |
    #                 | UNINITIALIZED  |
    #
    #                 ...
    #                 | UNINITIALIZED  |
    #                 +----------------+
    #
    # notes:
    # * rows correspond to a single individual in the evolutionary simulation
    #   (i.e., an occupant of a particular population position at a single
    #    generation)
    # * parentage buffer works like a std::vector in C++ in the y (# rows)
    #   direction: extra, unused rows are allocated (but rows past num_records
    #   are empty); when space runs out, the buffer row count is grown 2x
    # * number of rows for each generation will decrease below population_size
    #   as rows corresponding to individuals without extant ancestors are
    #   garbage collected
    # * self-loop parent_row_idx's denotes no parent (i.e., lineage begin)

    _loc_buffer: np.array  # [int]
    """Population position of each tracked organism; entries describe same
    organism as entry with corresponding row index in `_parentage_buffer`."""

    _trait_buffer: np.array  # [float]
    """Phenotypic trait of each tracked organism; entries describe same
    organism as entry with corresponding row index in `_parentage_buffer`."""

    def __init__(
        self: "GarbageCollectingPhyloTracker",
        initial_population: typing.Union[int, np.array],  # [float]
        working_buffer_size: typing.Optional[int] = None,
        share_common_ancestor: bool = True,
    ) -> None:
        """Initialize data structure to perfectly track phylogenetic history of
        a population.

        Parameters
        ----------
        initial_population : int or numpy array of float
            Specification of founding organisms of population. Providing an
            integer arguent specifices population size, and founding organisms'
            float phenotypic traits are recorded as NaN. Providing a numpy
            array of float specifies phenotypic traits of founding organisms.
        working_buffer_size : int, optional
            Intermittent garbage collection depth.

            How many rows back should intermittent garbage collection reach?
            Adjustments to this parameter only affect performance, not tracking
            semantics. Initial buffer size is set to 150% of
            `working_buffer_size`.
        share_common_ancestor : bool, default True
            Should a dummy common ancestor be inserted as the first entry in
            the tracker?

            If True, all initial population members will be recorded as
            children of this dummy ancestor. If False, all initial
            population members will be recorded as having no parent.

            The dummy ancestor's trait value is recorded as NaN and population
            loc is recorded as 0.
        """

        if isinstance(initial_population, int):
            initial_population = np.full(initial_population, np.nan)

        self._population_size = len(initial_population)
        assert self._population_size

        if working_buffer_size is None:
            working_buffer_size = 1000 * self._population_size

        self._working_buffer_size = working_buffer_size
        self._working_buffer_end = working_buffer_size

        assert (
            working_buffer_size
            >= self._population_size + share_common_ancestor
        )

        initial_buffer_size = working_buffer_size * 3 // 2
        self._parentage_buffer = np.empty(
            initial_buffer_size,
            dtype=np.int32,
        )
        self._loc_buffer = np.empty(
            initial_buffer_size,
            dtype=np.min_scalar_type(self._population_size),
        )
        self._trait_buffer = np.empty(
            initial_buffer_size,
            dtype=np.single,
        )

        self._num_records = self._population_size + share_common_ancestor

        if share_common_ancestor:
            self._parentage_buffer[0] = 0
            self._loc_buffer[0] = 0
            self._trait_buffer[0] = np.nan

            self._parentage_buffer[1 : self._population_size + 1] = 0
            self._loc_buffer[1 : self._population_size + 1] = np.arange(
                self._population_size
            )
            self._trait_buffer[
                1 : self._population_size + 1
            ] = initial_population
        else:
            self._parentage_buffer[: self._population_size] = np.arange(
                self._population_size
            )
            self._loc_buffer[: self._population_size] = np.arange(
                self._population_size
            )
            self._trait_buffer[: self._population_size] = initial_population

    def _GetBufferCapacity(self: "GarbageCollectingPhyloTracker") -> int:
        """How much buffer space is allocated?"""
        return self._parentage_buffer.shape[0]

    def _GarbageCollect(
        self: "GarbageCollectingPhyloTracker",
        below_row: int = 0,
    ) -> None:
        """Prune entries with extinct lineages at or after `below_row`."""
        assert self._num_records > below_row

        referenced_rows = _discern_referenced_rows(
            self._parentage_buffer,
            self._num_records,
            self._population_size,
            below_row,
        )
        num_referenced_rows = np.sum(referenced_rows)
        cumsum = np.cumsum(np.invert(referenced_rows))
        targets = (
            self._parentage_buffer[below_row : self._num_records] - below_row
        )
        if below_row:
            adjustment = cumsum[np.maximum(targets, 0)]
            adjustment[targets < 0] = 0
        else:
            adjustment = cumsum[targets]
        self._parentage_buffer[below_row : self._num_records] -= adjustment

        # condense non-garbage-collected rows to front of buffer
        self._parentage_buffer[
            below_row : below_row + num_referenced_rows
        ] = self._parentage_buffer[below_row : self._num_records][
            referenced_rows
        ]
        self._trait_buffer[
            below_row : below_row + num_referenced_rows
        ] = self._trait_buffer[below_row : self._num_records][referenced_rows]
        self._loc_buffer[
            below_row : below_row + num_referenced_rows
        ] = self._loc_buffer[below_row : self._num_records][referenced_rows]

        self._num_records = num_referenced_rows + below_row

        self._working_buffer_end = (
            self._num_records + 2 * self._working_buffer_size // 3
        )

    def _GarbageCollectWorkingBuffer(
        self: "GarbageCollectingPhyloTracker",
    ) -> bool:
        """Prune entries with extinct lineages within the fixed-size working
        buffer segment."""
        self._GarbageCollect(
            max(self._num_records - self._working_buffer_size, 0)
        )

    def _WouldInsertionOverflow(
        self: "GarbageCollectingPhyloTracker", num_to_insert: int
    ) -> bool:
        """Is buffer capacity sufficinet to accomodate `num_to_insert`
        insertions?"""
        return self._num_records + num_to_insert >= self._GetBufferCapacity()

    def _GrowBuffer(self: "GarbageCollectingPhyloTracker") -> None:
        """Allocate a additional buffer space.

        Buffer grows by a fixed proportion of current buffer size.
        """
        # refcheck only seems to fail in coverage build,
        # but set False everywhere anyways
        self._parentage_buffer.resize(
            int(self._GetBufferCapacity() * 1.5),
            refcheck=False,
        )
        self._trait_buffer.resize(
            int(self._GetBufferCapacity() * 1.5),
            refcheck=False,
        )
        self._loc_buffer.resize(
            int(self._GetBufferCapacity() * 1.5),
            refcheck=False,
        )

    def _GrowBufferForInsertion(
        self: "GarbageCollectingPhyloTracker", num_to_insert: int
    ) -> None:
        """Allocate sufficient additonal buffer space to accomodate
        `num_to_insert` insertions."""
        while self._WouldInsertionOverflow(num_to_insert):
            self._GrowBuffer()

    def ElapseGeneration(
        self: "GarbageCollectingPhyloTracker",
        parent_locs: typing.List[int],  # np.array ok
        traits: typing.Optional[typing.List[float]] = None,  # np.array ok
    ) -> None:
        """Append generational turnover to the phylogenetic record.

        Parameters
        ----------
        parent_locs : array_like of int
            Parent's population loc for each population loc.

            Position within array corresponds to post-turnover population
            members' population positions. Values within array correspond to
            those members' parents' population positions within the pre-
            turnover population.
        traits: array_like of float, optional
            Traits of post-turnover population members.

            If unspecified, post-turnover population members' traits are
            recorded as NaN.
        """
        assert self._population_size == len(parent_locs)

        if self._num_records >= self._working_buffer_end:
            self._GarbageCollectWorkingBuffer()

        if self._WouldInsertionOverflow(self._population_size):
            self._GrowBufferForInsertion(self._population_size)
            self._GarbageCollect()

        begin_row = self._num_records
        end_row = begin_row + self._population_size
        self._parentage_buffer[begin_row:end_row] = (
            begin_row - self._population_size + parent_locs
        )
        self._loc_buffer[begin_row:end_row] = np.arange(self._population_size)
        if traits is not None:
            self._trait_buffer[begin_row:end_row] = traits

        self._num_records = end_row

    def ApplyLocSwaps(
        self: "GarbageCollectingPhyloTracker",
        swapfrom_locs: np.array,  # [int]
        swapto_locs: np.array,  # [int]
    ) -> None:
        """Swap organisms between population locations.

        Organisms' pre-swap population locations do not remain in the
        phylogenetic record. Zip of parameters specifies pairs of population
        locations to swap.

        Notes
        -----
        Swaps are performed sequentially, with the most recent organism swapped
        to a location used for subsequent swaps from or to that location.

        Parameter order is commutative.
        """
        begin_row = self._num_records - self._population_size
        end_row = self._num_records
        apply_swaps(
            self._parentage_buffer[begin_row:end_row],
            swapfrom_locs,
            swapto_locs,
        )
        apply_swaps(
            self._trait_buffer[begin_row:end_row], swapfrom_locs, swapto_locs
        )

    def ApplyLocPasteovers(
        self: "GarbageCollectingPhyloTracker",
        copyfrom_locs: np.array,  # [int]
        copyto_locs: np.array,  # [int]
    ) -> None:
        """Replace organisms at `copyfto_locs` locations with non-descendant
        clones of organisms at `copyfrom_locs` locations.

        Pasted-over organisms do not remain in the phylogenetic record. The
        pasted-in organism is not recorded as the child of the pasted-from
        organism; rather, the pasted-in organism is recorded as the child of
        the pasted-from organism's parent.

        Zip of parameters specifies pasteovers to perform, with the first
        element of zipped tuples specifying the population location to copy
        from and the second element specifying the population location to paste
        over.

        Notes
        -----
        The original organism at each location is copied from, ignoring any
        preceding pasteovers that may have placed a new organism at that
        location during this operation.

        No repeated entries are allosed in `copyto_locs`.
        """
        assert len(copyto_locs) == count_unique(copyto_locs)

        begin_row = self._num_records - self._population_size
        end_row = self._num_records
        self._parentage_buffer[begin_row:end_row][
            copyto_locs
        ] = self._parentage_buffer[begin_row:end_row][copyfrom_locs].copy()
        self._trait_buffer[begin_row:end_row][
            copyto_locs
        ] = self._trait_buffer[begin_row:end_row][copyfrom_locs].copy()

    def CompilePhylogeny(
        self: "GarbageCollectingPhyloTracker",
        progress_wrap=lambda x: x,
        loc_transforms: typing.Dict[
            str, typing.Callable
        ] = types.MappingProxyType({}),
    ) -> pd.DataFrame:
        """Generate full phylogenetic record of extant organisms.

        Parameters
        ----------
        progress_wrap : Callable, default identity function
            Wrapper applied around record row iterator; pass tqdm or equivalent
            to display progress bar for compilation process.
        loc_transforms: Dict of str -> Callable, default empty
            Specification for custom columns in returned dataframe.

            Each dict item creates a new column with name corresponding to the
            item's key populated with the result of mapping the item's Callable value over organisms' locs.

        Returns
        -------
        pandas.DataFrame
            Full phylogenetic record of extant organisms alife standard format.

        Notes
        -----
        This operation is non-destructive; further record-keeping may be
        performed on the tracker object afterwards.
        """
        self._GarbageCollect()
        generation_lookup = collections.Counter()

        records = []
        for idx, (parent_idx, loc, trait) in enumerate(
            zip(
                progress_wrap(self._parentage_buffer[: self._num_records]),
                self._loc_buffer[: self._num_records],
                self._trait_buffer[: self._num_records],
            )
        ):
            generation_lookup[idx] = generation_lookup[parent_idx] + (
                parent_idx != idx
            )
            records.append(
                {
                    **{
                        "id": idx,
                        "ancestor_list": str(
                            [parent_idx if parent_idx != idx else None]
                        ),
                        "loc": loc,
                        "trait": trait,
                        "origin_time": generation_lookup[idx],
                    },
                    **{
                        key: transform(loc)
                        for key, transform in loc_transforms.items()
                    },
                }
            )
        return pd.DataFrame.from_records(records)



"""Frozen representations of genome annotations for efficient postprocessing
and analysis."""

from ._HereditaryStratigraphicAssemblage import (
    HereditaryStratigraphicAssemblage,
)
from ._HereditaryStratigraphicAssemblageSpecimen import (
    HereditaryStratigraphicAssemblageSpecimen,
)
from ._HereditaryStratigraphicSpecimen import HereditaryStratigraphicSpecimen

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "HereditaryStratigraphicAssemblage",
    "HereditaryStratigraphicAssemblageSpecimen",
    "HereditaryStratigraphicSpecimen",
]

from .._auxiliary_lib import launder_impl_modules as _launder

_launder(
    [
        HereditaryStratigraphicAssemblage,
        HereditaryStratigraphicAssemblageSpecimen,
        HereditaryStratigraphicSpecimen,
    ],
    __name__,
)
del _launder  # prevent name from leaking



import functools
import typing

import numpy as np
import pandera as pa

from .._auxiliary_lib import (
    CopyableSeriesItemsIter,
    get_nullable_mask,
    get_nullable_vals,
)

_nullable_unsigned_integer_series_t = typing.Union[
    pa.typing.Series[pa.typing.UINT8()],
    pa.typing.Series[pa.typing.UINT16()],
    pa.typing.Series[pa.typing.UINT32()],
    pa.typing.Series[pa.typing.UINT64()],
]


class HereditaryStratigraphicAssemblageSpecimen:
    """Postprocessing representation of the differentia retained by an extant
    HereditaryStratigraphicColumn, indexed by deposition rank.

    Differentia are stored using a nullable integer representation, which allows
    for inclusion of entries for all ranks retained by any specimen within the
    assemblage, even if that particualr rank is not retained by this specimen.
    This allows for more efficient comparisons between specimens, due to direct
    alignment.

    See Also
    --------
    HereditaryStratigraphicSpecimen
        Specimen representation that can contain only ranks retained by that
        specimen.
    HereditaryStratigraphicAssemblage
        Gathers a collection of `HereditaryStratigraphicSpecimen`s and
        facilitates creation of corresponding aligned `HereditaryStratigraphicAssemblageSpecimen`s.
    """

    __slots__ = ("_data", "_stratum_differentia_bit_width")

    _data: _nullable_unsigned_integer_series_t
    _stratum_differentia_bit_width: int

    def __init__(
        self: "HereditaryStratigraphicAssemblageSpecimen",
        stratum_differentia_series: _nullable_unsigned_integer_series_t,
        stratum_differentia_bit_width: int,
    ) -> None:
        """Initialize a HereditaryStratigraphicAssemblageSpecimen object with a
        (potentially sparse) sequence of rank-indexed differentia and a
        differentia bit width."""
        self._data = stratum_differentia_series
        self._stratum_differentia_bit_width = stratum_differentia_bit_width

    def GetStratumDifferentiaBitWidth(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> int:
        """How many bits wide are the differentia of strata?"""
        return self._stratum_differentia_bit_width

    @functools.lru_cache(maxsize=None)
    def GetNumStrataDeposited(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> int:
        """How many strata have been depostited on the column?

        Note that a first stratum is deposited on the column during
        initialization.
        """
        return self._data.last_valid_index() + 1

    @functools.lru_cache(maxsize=None)
    def GetNumStrataRetained(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> int:
        """How many strata are currently stored within the column?

        May be fewer than the number of strata deposited if strata have been
        discarded as part of the configured stratum retention policy.
        """
        return len(self._data) - self.GetStratumMask().sum()

    def GetNumDiscardedStrata(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> int:
        """How many deposited strata have been discarded?

        Determined by number of generations elapsed and the configured column
        retention policy.
        """
        return self.GetNumStrataDeposited() - self.GetNumStrataRetained()

    def HasDiscardedStrata(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> bool:
        """Have any deposited strata been discarded?"""
        return self.GetNumDiscardedStrata() > 0

    def GetData(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> _nullable_unsigned_integer_series_t:
        """Get the underlying Pandas Series containing differentia values
        indexed by rank.

        Notes
        -----
        This function directly returns the specimen's underlying Series
        data, so mutation of the returned object will alter or invalidate this
        specimen.
        """
        return self._data

    def GetStratumMask(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> np.ndarray:
        """Get a boolean mask indicating which entries in the stored Pandas
        NullableInteger Series are null.

        I.e., which ranks does this specimen not retain differentia at?

        Returns
        -------
        mask : np.ndarray
            A 1-dimensional boolean NumPy ndarray.

            True values indicate nullness.

        Notes
        -----
        This function returns the underlying boolean mask used by the stored
        Pandas Series object to represent null values. This mask is a direct
        view into the Series data, so no copy is made. Changes to the mask will
        propagate to the store Series object, and vice versa.
        """
        try:
            return get_nullable_mask(self._data)
        except AttributeError:  # object type (not nullable int) array
            return self._data.isna()

    def GetDifferentiaVals(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> np.ndarray:
        """Get the integer underlying values in the stored Pandas
        NullableInteger Series.

        Returns
        -------
        differentia : np.ndarray
            A 1-dimensional NumPy integer ndarray containing differentia values,
            including garbage values where the underlying Series is null.

        Notes
        -----
        This function returns a direct view into the Series data, so no copy is
        made. Changes to the returned array will propagate to the Series
        object's underlying values, and vice versa.
        """
        try:
            return get_nullable_vals(self._data)
        except AttributeError:  # object type (not nullable int) array
            return self._data.array

    def GetRankIndex(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> np.ndarray:
        """Get the integer index in the stored Pandas Series, representing the
        ranks of stratum entries.

        Returns
        -------
        ranks : np.ndarray
            A numpy array containing ranks of differentia entries, including
            null entries for differentia that are not retained.
        """
        return self._data.index.array.to_numpy()

    def GetRankAtColumnIndex(
        self: "HereditaryStratigraphicAssemblageSpecimen",
        index: int,
    ) -> int:
        """Map array position to generation of deposition.

        What is the deposition rank of the stratum positioned at index i
        among retained strata? Index order is from most ancient (index 0) to
        most recent.
        """
        return self.GetData().dropna().index[index]

    def IterRetainedRanks(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> typing.Iterator[int]:
        """Iterate over deposition ranks of strata retained in the specimen."""
        yield from self.GetData().dropna().index

    def IterRetainedDifferentia(
        self: "HereditaryStratigraphicAssemblageSpecimen",
    ) -> typing.Iterator[int]:
        """Iterate over differentia of strata retained in the specimen.

        Differentia yielded from most ancient to most recent.
        """
        yield from self.GetData().dropna()

    def IterRankDifferentiaZip(
        self: "HereditaryStratigraphicAssemblageSpecimen",
        copyable: bool = False,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over ranks of retained strata and their differentia.

        If `copyable`, return an iterator that can be copied to produce a new
        fully-independent iterator at the same position.

        Equivalent to `zip(specimen.IterRetainedRanks(),
        specimen.IterRetainedDifferentia())`, but may be more efficient.
        """
        if copyable:
            return CopyableSeriesItemsIter(self._data.dropna())
        else:
            return self._data.dropna().items()



import functools
import typing

from ..._detail import PolicyCouplerBase
from ._iter_priority_ranks import iter_priority_ranks


@functools.lru_cache(maxsize=512)
def get_retained_ranks(
    policy: PolicyCouplerBase,
    num_strata_deposited: int,
) -> typing.Set[int]:
    """Calculate the set of strata ranks retained at `num_strata_deposited`."""
    spec = policy.GetSpec()
    degree = spec.GetDegree()
    interspersal = spec.GetInterspersal()

    # special case
    if num_strata_deposited == 0:
        return set()

    last_rank = num_strata_deposited - 1
    # we will always retain the zeroth rank and the last rank
    # Set data structure prevents duplicates
    res = {0, last_rank}

    # create iterators that yield ranks for retention in priority order for
    # each target recency
    iters = [
        iter_priority_ranks(degree, interspersal, pow, num_strata_deposited)
        # note:
        # even though 0th pow is always just the most recent rank
        # we need to iterate over it because it will eventually yield
        # all preceding ranks ensuring that we fill available space
        # HOWEVER, it is excluded from the first round and is only drawn
        # from subsequently to ensure that it will have lowest priority
        # thereby making optimizations easier [and requiring less space
        # be devoted to the equivalent of
        # reversed(range(num_strata_deposited))]
        for pow in reversed(range(1, degree + 1))
    ]
    # round robin, taking at least one rank from each iterator until the
    # upper bound on space complexity is exactly reached or all iterators
    # are exhausted
    while len(res) < policy.CalcNumStrataRetainedUpperBound(
        num_strata_deposited,
    ):
        res_before = len(res)
        for iter_ in iters:
            # will loop 0 times if iter_ is empty
            for priority_rank in iter_:
                # draw from iter_ until a rank not already in res is
                # discovered or iter_ is exhausted
                if priority_rank not in res:
                    res.add(priority_rank)
                    break
            # ensure space complexity limit is not exceeded
            if len(res) == policy.CalcNumStrataRetainedUpperBound(
                num_strata_deposited,
            ):
                break
        # if no progress was made then all iter_ were empty
        # and its time to quit
        if res_before == len(res):
            break

    # draw from pow 0 iter until res full
    # (pow 0 iter only drawn from if all other iters are exhausted)
    for priority_rank in iter_priority_ranks(
        degree, interspersal, 0, num_strata_deposited
    ):
        # ensure space complexity limit is not exceeded
        if len(res) == policy.CalcNumStrataRetainedUpperBound(
            num_strata_deposited,
        ):
            break
        res.add(priority_rank)

    # sanity checks then return
    assert all(0 <= n < num_strata_deposited for n in res)
    assert len(res) <= policy.CalcNumStrataRetainedUpperBound(
        num_strata_deposited,
    )
    assert len(res) == policy.CalcNumStrataRetainedExact(num_strata_deposited)
    assert res
    if len(res) < policy.CalcNumStrataRetainedUpperBound(
        num_strata_deposited,
    ):
        assert res == {*range(len(res))}

    return res



import operator
import unittest

from hstrat._auxiliary_lib import find_bounds


class TestFindBounds(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        assert find_bounds(
            query=5,
            iterable=iter(()),
        ) == (None, None)

    def test_singleton_above(self):
        assert find_bounds(
            query=5,
            iterable=(7,),
        ) == (None, 7)

    def test_singleton_below(self):
        assert find_bounds(
            query=5,
            iterable=(0,),
        ) == (0, None)

    def test_singleton_equal(self):
        assert find_bounds(
            query=5,
            iterable=(5,),
        ) == (None, None)

    def test_doublet_equal(self):
        assert find_bounds(
            query=5,
            iterable=(5, 5),
        ) == (None, None)

    def test_doublet_above(self):
        assert find_bounds(
            query=5,
            iterable=(8, 7),
        ) == (None, 7)

    def test_doublet_below(self):
        assert find_bounds(
            query=5,
            iterable=(-10, -1),
        ) == (-1, None)

    def test_doublet_split(self):
        assert find_bounds(
            query=5,
            iterable=(8, -1),
        ) == (-1, 8)
        assert find_bounds(
            query=5,
            iterable=(-1, 8),
        ) == (-1, 8)

    def test_long_with_edgecases(self):
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
        ) == (3, 6)

    def test_key(self):
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
            key=lambda x: -x,
        ) == (1, 10)

    def test_initializer(self):
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
            initializer=(None, 7),
        ) == (3, 6)
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
            initializer=(4, 7),
        ) == (4, 6)
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
            initializer=(4, None),
        ) == (4, 6)
        assert find_bounds(
            query=5, iterable=(8, 8, 5, 7, 20, 6, 100), initializer=(4, None)
        ) == (4, 6)

    def test_filter_above(self):
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
            filter_above=operator.ge,
        ) == (3, 5)

    def test_filter_below(self):
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
            filter_below=operator.le,
        ) == (5, 6)

    def test_filter_abovebelow(self):
        assert find_bounds(
            query=5,
            iterable=(8, 8, 5, -1, 7, -1, -10, 0, 5, 5, 20, 3, 3, 6, 100),
            filter_above=operator.ge,
            filter_below=operator.le,
        ) == (5, 5)


if __name__ == "__main__":
    unittest.main()



from ....._auxiliary_lib import bit_floor
from ._calc_target_recency import calc_target_recency


def calc_rank_sep(
    degree: int,
    interspersal: int,
    pow_: int,
    num_strata_deposited: int,
) -> int:
    """How far apart should ranks retained to cover the `pow`th target recency
    be spaced?

    Will be a power of 2 monotonically increasing with `num_strata_deposited`.
    """
    target_recency = calc_target_recency(degree, pow_, num_strata_deposited)
    # spacing between retained ranks
    target_retained_ranks_sep = max(
        target_recency / interspersal,
        1.0,
    )
    # round down to power of 2
    retained_ranks_sep = bit_floor(int(target_retained_ranks_sep))
    return retained_ranks_sep



from collections import Counter
import functools
import itertools as it
import os
import random

from Bio import Phylo as BioPhylo
import alifedata_phyloinformatics_convert as apc
import dendropy as dp
import pandas as pd
import pytest
from tqdm import tqdm

from hstrat import hstrat

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "always_store_rank_in_stratum",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "num_predeposits",
    [
        0,
        1,
        10,
        100,
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.pseudostochastic_algo.Policy(hash_salt=3),
        hstrat.depth_proportional_resolution_algo.Policy(
            depth_proportional_resolution=10
        ),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "tree_path",
    [
        "",
        f"{assets_path}/grandchild_and_aunt.newick",
        f"{assets_path}/grandchild_and_auntuncle.newick",
        f"{assets_path}/grandchild.newick",
        f"{assets_path}/grandtriplets_and_aunt.newick",
        f"{assets_path}/grandtriplets_and_auntuncle.newick",
        f"{assets_path}/grandtriplets.newick",
        f"{assets_path}/grandtwins_and_aunt.newick",
        f"{assets_path}/grandtwins_and_auntuncle.newick",
        f"{assets_path}/grandtwins.newick",
        f"{assets_path}/justroot.newick",
        f"{assets_path}/triplets.newick",
        f"{assets_path}/twins.newick",
        f"{assets_path}/nk_tournamentselection.csv",
    ],
)
def test_descend_template_phylogeny(
    always_store_rank_in_stratum,
    num_predeposits,
    retention_policy,
    tree_path,
):

    if tree_path == "":
        tree = dp.Tree()
        bp_tree = BioPhylo.BaseTree.Tree()
    elif tree_path.endswith(".newick"):
        tree = dp.Tree.get(path=tree_path, schema="newick")
        bp_tree = BioPhylo.read(tree_path, "newick")
    elif tree_path.endswith(".csv"):
        tree = apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(tree_path),
        )
        bp_tree = apc.alife_dataframe_to_biopython_tree(
            pd.read_csv(tree_path),
        )
    else:
        assert False

    for node in tree:
        node.edge_length = 1
    for clade in bp_tree.find_clades():
        clade.branch_length = 1
    tree.seed_node.edge.length = num_predeposits
    bp_tree.rooted = True
    bp_tree.root.branch_length = num_predeposits

    for loc, node in enumerate(tree.leaf_node_iter()):
        node.taxon = tree.taxon_namespace.new_taxon(label=str(loc))

    tree.update_bipartitions(
        suppress_unifurcations=False,
        collapse_unrooted_basal_bifurcation=False,
    )

    # setup seed column
    seed_column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_differentia_bit_width=64,
    )
    seed_column.DepositStrata(num_stratum_depositions=num_predeposits)

    extant_population = hstrat.descend_template_phylogeny_biopython(
        bp_tree,
        seed_column=seed_column,
        progress_wrap=functools.partial(tqdm, disable=True),
    )

    num_tips = len(tree)
    assert num_tips == len(extant_population)

    tip_depths = [
        int(tip_node.distance_from_root())
        for tip_node in tree.leaf_node_iter()
    ]
    assert tip_depths == [
        column.GetNumStrataDeposited() - 1 for column in extant_population
    ]

    extant_population_dp = hstrat.descend_template_phylogeny_dendropy(
        tree,
        seed_column=seed_column,
        progress_wrap=functools.partial(tqdm, disable=True),
    )

    assert len(extant_population) == len(extant_population_dp)

    assert Counter(
        hstrat.calc_rank_of_mrca_bounds_between(c1, c2, prior="arbitrary")
        for c1, c2 in it.permutations(extant_population, 2)
    ) == Counter(
        hstrat.calc_rank_of_mrca_bounds_between(c1, c2, prior="arbitrary")
        for c1, c2 in it.permutations(extant_population_dp, 2)
    )

    extant_nodes = random.sample(
        [*bp_tree.find_clades()], min(2, len([*bp_tree.find_clades()]))
    )
    extant_population = hstrat.descend_template_phylogeny_biopython(
        bp_tree,
        seed_column=seed_column,
        extant_nodes=extant_nodes,
        progress_wrap=functools.partial(tqdm, disable=True),
    )
    assert len(extant_nodes) == len(extant_population)

    depths = bp_tree.depths()
    assert Counter(depths[n] for n in extant_nodes) == Counter(
        c.GetNumStrataDeposited() - 1 for c in extant_population
    )



import pytest

from hstrat._auxiliary_lib import (
    assign_intersecting_subsets,
    coshuffled,
    make_intersecting_subsets,
)


@pytest.mark.parametrize(
    "ranges, expected_subsets_indices, expected_subsets_ranges",
    [
        (
            [(0, 1), (2, 3), (4, 5)],
            [[0], [1], [2]],
            [[(0, 1)], [(2, 3)], [(4, 5)]],
        ),
        (
            [(0, 2), (1, 2), (2, 3)],
            [[0, 1], [2]],
            [[(0, 2), (1, 2)], [(2, 3)]],
        ),
        (
            [(0, 5), (1, 2), (2, 3)],
            [[0, 1], [2]],
            [[(0, 5), (1, 2)], [(2, 3)]],
        ),
        (
            [(0, 5), (1, 2), (2, 3), (3, 3)],
            [[0, 1], [2], [3]],
            [[(0, 5), (1, 2)], [(2, 3)], [(3, 3)]],
        ),
        ([(0, 1)], [[0]], [[(0, 1)]]),
        ([], [], []),
        (
            [(0, 2), (1, 2), (2, 3)],
            [[0, 1], [2]],
            [[(0, 2), (1, 2)], [(2, 3)]],
        ),
        (
            [(0, 2), (1, 2), (2, 2)],
            [[0, 1], [2]],
            [[(0, 2), (1, 2)], [(2, 2)]],
        ),
        (
            [(0, 1), (0, 1)],
            [[0, 1]],
            [[(0, 1), (0, 1)]],
        ),
        (
            [(0, 3), (1, 2), (1, 3)],
            [[0, 1, 2]],
            [[(0, 3), (1, 2), (1, 3)]],
        ),
        (
            [(1, 2), (2, 5), (2, 5), (0, 4), (3, 4)],
            [[0, 3], [1, 2, 4]],
            [[(0, 4), (1, 2)], [(2, 5), (2, 5), (3, 4)]],
        ),
    ],
)
def test_assign_intersecting_subsets(
    ranges, expected_subsets_indices, expected_subsets_ranges
):

    range_indices = [*range(len(ranges))]

    for _rep in range(20):
        created_subsets_ranges = make_intersecting_subsets(ranges)
        created_subsets_indices = assign_intersecting_subsets(ranges)
        for (
            created_subset_indices,
            created_subset_ranges,
            expected_subset_indices,
            expected_subset_ranges,
        ) in zip(
            created_subsets_indices,
            created_subsets_ranges,
            expected_subsets_indices,
            expected_subsets_ranges,
        ):

            assert len(created_subset_indices) == len(created_subset_ranges)
            assert len(created_subset_indices) == len(expected_subset_indices)

            assert sorted(
                [range_indices[i] for i in created_subset_indices]
            ) == sorted(expected_subset_indices) or len(ranges) > len(
                set(ranges)
            )

            recreated_subset_ranges = [
                ranges[i] for i in created_subset_indices
            ]
            assert (
                sorted(recreated_subset_ranges)
                == sorted(expected_subset_ranges)
                == sorted(created_subset_ranges)
            )

        ranges, range_indices = coshuffled(ranges, range_indices)



import itertools as it
import typing

from .._auxiliary_lib import get_hstrat_version, log_once_in_a_row
from ..genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratumOrderedStoreList,
)
from ._impl import policy_from_record
from ._unpack_differentiae_str import unpack_differentiae_str


def col_from_records(records: typing.Dict) -> HereditaryStratigraphicColumn:
    """Deserialize a `HereditaryStratigraphicColumn` from a dict composed of
    builtin data types.
    """
    if get_hstrat_version() != records["hstrat_version"]:
        log_once_in_a_row(
            f"""col_from_records version mismatch, record is version {
                records['hstrat_version']
            } and software is version {
                get_hstrat_version()
            }"""
        )

    policy = policy_from_record(records["policy"])

    def load_stratum_ordered_store() -> HereditaryStratumOrderedStoreList:
        dummy_column = HereditaryStratigraphicColumn(
            stratum_retention_policy=policy_from_record(records["policy"]),
            stratum_differentia_bit_width=records["differentia_bit_width"],
            always_store_rank_in_stratum=(
                "stratum_deposition_ranks" in records
            ),
        )
        store = HereditaryStratumOrderedStoreList()

        for differentia, deposition_rank, annotation in zip(
            unpack_differentiae_str(
                records["differentiae"],
                differentia_bit_width=records["differentia_bit_width"],
                num_packed_differentia=(
                    None
                    if not records.get("omits_num_padding_bits_header", False)
                    else policy.CalcNumStrataRetainedExact(
                        records["num_strata_deposited"],
                    )
                ),
            ),
            records.get("stratum_deposition_ranks", it.repeat(None)),
            records.get("stratum_annotations", it.repeat(None)),
        ):
            store.DepositStratum(
                rank=deposition_rank,
                stratum=dummy_column._CreateStratum(
                    annotation=annotation,
                    differentia=differentia,
                    deposition_rank=deposition_rank,
                ),
            )
        return store

    return HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=records["differentia_bit_width"],
        stratum_ordered_store=(
            load_stratum_ordered_store(),
            records["num_strata_deposited"],
        ),
        always_store_rank_in_stratum=("stratum_deposition_ranks" in records),
    )



import itertools as it
import typing

import numpy as np


def estimate_rank_of_mrca_naive(
    coincident_ranks_from_first_disparity_through_first_commonality: typing.Iterator[
        int
    ],
    p_differentia_collision: float,
    prior: object,
) -> typing.Optional[float]:
    """Compute a simple, fast estimate the rank of the most recent common
    ancestor (MRCA).

    Parameters
    ----------
    coincident_ranks_from_first_disparity_through_first_commonality
    : typing.Iterator[int]
        Iterator of integer values indicating the coincident ranks between two
        taxa.

        Inclusive to first disparity and first commonality.
    p_differentia_collision : float
        The multiplicative inverse of the number of possible differentia.
    prior : object
        Prior expectation for the distribution of MRCA generation
        between hereditary stratigraphic columns/

    Returns
    -------
    typing.Optional[float]
        Estimated rank of the MRCA, or None if the two hereditary stratigraphic
        artifacts definitvely share no common ancestor.

    Notes
    -----
    This function estimates the rank of the MRCA by computing the mean of the
    first retained disparite rank and the last retained common rank, with a correction factor for upper bound exclusivity.
    """

    rank_of_mrca_bounds = tuple(
        it.islice(
            coincident_ranks_from_first_disparity_through_first_commonality,
            2,
        )
    )
    assert len(rank_of_mrca_bounds)
    if len(rank_of_mrca_bounds) == 1:
        # rank 0 is a disparity --- definitively share no common ancestor
        return None

    exclusive_ub_correction = 1 / 2
    return np.mean(rank_of_mrca_bounds) - exclusive_ub_correction



import unittest

from hstrat._auxiliary_lib import div_range


class TestDivRange(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):

        assert [*div_range(10, 12, 2)] == []
        assert [*div_range(10, 10, 2)] == []

    def test_singleton(self):

        assert [*div_range(10, 8, 2)] == [10]

    def test_several(self):

        assert [*div_range(10, 1, 2)] == [10, 5, 2]


if __name__ == "__main__":
    unittest.main()



import sys


# adapted from https://stackoverflow.com/a/44595269
def is_in_unit_test() -> bool:
    """Is current execution within a unit testing context?

    Can be used to enable more thorough assert checks than desired in normal
    (i.e., non-optimized) mode outside testing context.
    """
    return "pytest" in sys.modules



import typing

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_coarsen_mask,
    alifestd_make_empty,
    alifestd_validate,
)


def make_asexual_phylo_df() -> pd.DataFrame:
    """Creates an example dataframe with 7 rows."""
    return pd.DataFrame(
        {
            "id": [*range(7)],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
            "ancestor_id": [0, 0, 0, 1, 1, 2, 2],
        }
    )


@pytest.fixture  # wrapper because user cannot call fixtures directly
def asexual_phylo_df() -> pd.DataFrame:
    return make_asexual_phylo_df()


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("drop_ancestor_id", [True, False])
@pytest.mark.parametrize(
    "reorder",
    [
        None,
        # roll first row to end to make topo unsorted
        lambda x: x.apply(np.roll, shift=1),
    ],
)
@pytest.mark.parametrize(
    "mask, expected_df",
    [
        # Everything is included in the mask
        (
            pd.Series([True] * 7),
            make_asexual_phylo_df(),
        ),
        # Root singleton dataframe
        (
            pd.Series([True, False, False, False, False, False, False]),
            pd.DataFrame(
                {"id": [0], "ancestor_list": ["[none]"], "ancestor_id": [0]}
            ),
        ),
        # Non-root singleton dataframe
        (
            pd.Series([False, True, False, False, False, False, False]),
            pd.DataFrame(
                {"id": [1], "ancestor_list": ["[none]"], "ancestor_id": [1]}
            ),
        ),
        # Nothing is included in the mask
        (
            pd.Series([False] * 7),
            alifestd_make_empty(ancestor_id=True),
        ),
        # Root is included in the mask
        (
            pd.Series([True, False, False, False, False, False, False]),
            pd.DataFrame(
                {"id": [0], "ancestor_list": ["[none]"], "ancestor_id": [0]}
            ),
        ),
        # Root is not included in the mask
        (
            pd.Series([False, True, True, True, True, True, True]),
            pd.DataFrame(
                {
                    "id": list(range(1, 7)),
                    "ancestor_list": [
                        "[none]",
                        "[none]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                    ],
                    "ancestor_id": [1, 2, 1, 1, 2, 2],
                },
            ),
        ),
        # No leaves are included in the mask
        (
            pd.Series([True, True, True, True, True, False, False]),
            pd.DataFrame(
                {
                    "id": [*range(5)],
                    "ancestor_list": ["[none]", "[0]", "[0]", "[1]", "[1]"],
                    "ancestor_id": [0, 0, 0, 1, 1],
                }
            ),
        ),
        # All leaves are included in the mask
        (
            pd.Series([False, False, False, False, False, True, True]),
            pd.DataFrame(
                {
                    "id": [5, 6],
                    "ancestor_list": ["[none]", "[none]"],
                    "ancestor_id": [5, 6],
                },
            ),
        ),
        # One internal node excluded from mask
        (
            pd.Series([True, True, False, True, True, True, True]),
            pd.DataFrame(
                {
                    "id": [0, 1, 3, 4, 5, 6],
                    "ancestor_list": [
                        "[none]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[0]",
                        "[0]",
                    ],
                    "ancestor_id": [0, 0, 1, 1, 0, 0],
                },
            ),
        ),
        # All internal nodes excluded from mask
        (
            pd.Series([True, False, False, True, True, True, False]),
            pd.DataFrame(
                {
                    "id": [0, 3, 4, 5],
                    "ancestor_list": ["[none]", "[0]", "[0]", "[0]"],
                    "ancestor_id": [0, 0, 0, 0],
                },
            ),
        ),
    ],
)
def test_asexual_phylo_df(
    asexual_phylo_df: pd.DataFrame,
    mask: pd.Series,
    expected_df: pd.DataFrame,
    drop_ancestor_id: bool,
    mutate: bool,
    reorder: typing.Optional[typing.Callable],
):
    if reorder is not None:
        asexual_phylo_df = reorder(asexual_phylo_df)
        mask = reorder(mask)
    if drop_ancestor_id:
        asexual_phylo_df = asexual_phylo_df.drop("ancestor_id", axis=1)
        expected_df = expected_df.drop("ancestor_id", axis=1)

    original_df = asexual_phylo_df.copy()
    result_df = alifestd_coarsen_mask(asexual_phylo_df, mask, mutate=mutate)
    assert alifestd_validate(result_df)
    if drop_ancestor_id:  # allow for addition of ancestor_id column
        result_df.drop("ancestor_id", axis=1, inplace=True)

    assert alifestd_validate(asexual_phylo_df)

    assert alifestd_validate(expected_df)
    if reorder is not None:
        result_df.sort_values(by="id", inplace=True, ignore_index=True)
        expected_df.sort_values(by="id", inplace=True, ignore_index=True)

    pd.testing.assert_frame_equal(
        result_df, expected_df, check_index_type=False, check_like=True
    )
    if not mutate:
        pd.testing.assert_frame_equal(
            asexual_phylo_df,
            original_df,
            check_index_type=False,
            check_like=True,
        )


def make_sexual_phylo_df() -> pd.DataFrame:
    """Creates an example dataframe with 7 rows representing a sexual phylogeny."""
    return pd.DataFrame(
        {
            "id": [*range(7)],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[0]",
                "[1, 2]",
                "[1, 2]",
                "[3]",
                "[3]",
            ],
        }
    )


@pytest.fixture  # wrapper because users cannot call fixtures directly
def sexual_phylo_df() -> pd.DataFrame:
    return make_sexual_phylo_df()


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize(
    "reorder",
    [
        None,
        # roll first row to end to make topo unsorted
        lambda x: x.apply(np.roll, shift=1),
    ],
)
@pytest.mark.parametrize(
    "mask, expected_df",
    [
        # Everything is included in the mask
        (
            pd.Series([True] * 7),
            make_sexual_phylo_df(),
        ),
        # Root singleton dataframe
        (
            pd.Series([True, False, False, False, False, False, False]),
            pd.DataFrame({"id": [0], "ancestor_list": ["[none]"]}),
        ),
        # Non-root singleton dataframe
        (
            pd.Series([False, True, False, False, False, False, False]),
            pd.DataFrame({"id": [1], "ancestor_list": ["[none]"]}),
        ),
        # A node with multiple ancestors
        (
            pd.Series([False, False, False, True, False, False, False]),
            pd.DataFrame({"id": [3], "ancestor_list": ["[none]"]}),
        ),
        # Exclude root, result has one ancestor for each entry except the first
        (
            pd.Series([False, True, True, True, True, True, True]),
            pd.DataFrame(
                {
                    "id": list(range(1, 7)),
                    "ancestor_list": [
                        "[none]",
                        "[none]",
                        "[1, 2]",
                        "[1, 2]",
                        "[3]",
                        "[3]",
                    ],
                },
            ),
        ),
        # Excluding one internal node and its descendants
        (
            pd.Series([True, True, True, False, False, False, False]),
            pd.DataFrame(
                {
                    "id": [*range(3)],
                    "ancestor_list": ["[none]", "[0]", "[0]"],
                }
            ),
        ),
        # Including only the leaves
        (
            pd.Series([False, False, False, False, False, True, True]),
            pd.DataFrame(
                {
                    "id": [5, 6],
                    "ancestor_list": ["[none]", "[none]"],
                },
            ),
        ),
        # Excluding internal nodes
        (
            pd.Series([True, False, False, False, False, True, True]),
            pd.DataFrame(
                {
                    "id": [0, 5, 6],
                    "ancestor_list": ["[none]", "[0]", "[0]"],
                },
            ),
        ),
        # One internal node excluded (with its descendants) from mask
        (
            pd.Series([True, True, False, True, True, False, False]),
            pd.DataFrame(
                {
                    "id": [0, 1, 3, 4],
                    "ancestor_list": [
                        "[none]",
                        "[0]",
                        "[0, 1]",
                        "[0, 1]",
                    ],
                },
            ),
        ),
    ],
)
def test_sexual_phylo_df(
    sexual_phylo_df: pd.DataFrame,
    mask: pd.Series,
    expected_df: pd.DataFrame,
    mutate: bool,
    reorder: typing.Optional[typing.Callable],
):
    if reorder is not None:
        sexual_phylo_df = reorder(sexual_phylo_df)
        mask = reorder(mask)
    original_df = sexual_phylo_df.copy()
    result_df = alifestd_coarsen_mask(sexual_phylo_df, mask, mutate=mutate)

    assert alifestd_validate(sexual_phylo_df)
    assert alifestd_validate(expected_df)
    if reorder is not None:
        result_df.sort_values(by="id", inplace=True, ignore_index=True)
        expected_df.sort_values(by="id", inplace=True, ignore_index=True)
    pd.testing.assert_frame_equal(
        result_df, expected_df, check_index_type=False, check_like=True
    )
    if not mutate:
        pd.testing.assert_frame_equal(
            sexual_phylo_df,
            original_df,
            check_index_type=False,
            check_like=True,
        )


@pytest.mark.parametrize("drop_ancestor_id", [True, False])
@pytest.mark.parametrize("mutate", [True, False])
def test_empty_df(drop_ancestor_id: bool, mutate: bool):
    original_df = alifestd_make_empty(ancestor_id=True)
    if drop_ancestor_id:
        original_df = original_df.drop("ancestor_id", axis=1)

    expected_df = original_df.copy()
    mask = pd.Series(dtype=bool)
    result_df = alifestd_coarsen_mask(original_df, mask, mutate=mutate)

    assert alifestd_validate(result_df)
    pd.testing.assert_frame_equal(
        result_df, expected_df, check_index_type=False, check_like=True
    )
    if not mutate:
        pd.testing.assert_frame_equal(
            original_df,
            original_df,
            check_index_type=False,
            check_like=True,
        )



"""Tools to create phylogenies and corresponding extant populations of
hereditary stratigraph columns."""

from . import (
    descend_template_phylogeny,
    generate_template_phylogeny,
    perfect_tracking,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = (
    descend_template_phylogeny.__all__
    + generate_template_phylogeny.__all__
    + perfect_tracking.__all__
)

from .descend_template_phylogeny import *  # noqa: F401
from .generate_template_phylogeny import *  # noqa: F401
from .perfect_tracking import *  # noqa: F401



import typing


def generate_n(generator: typing.Callable, n: int) -> typing.Iterator:
    """Returns an iterator that generates `n` values by repeatedly calling the given `generator` function.

    Parameters
    ----------
    generator : Callable
        A function that generates the values to be returned by the iterator. The function should take no arguments.
    n : int
        The number of values to generate.

    Returns
    -------
    iterator : Iterator
        An iterator that generates `n` values by repeatedly calling the `generator` function.

    Raises
    ------
    TypeError
        If `generator` is not callable.
    ValueError
        If `n` is negative.

    Examples
    --------
    >>> import random
    >>> random.seed(42)
    >>> def random_int():
    ...     return random.randint(1, 100)
    >>> values = list(generate_n(random_int, 5))
    >>> values
    [81, 14, 3, 94, 35]

    """
    for __ in range(n):
        yield generator()



"""Functors to configure policies that satisfy evaluated requirements."""

from ._PropertyAtLeastParameterizer import PropertyAtLeastParameterizer
from ._PropertyAtMostParameterizer import PropertyAtMostParameterizer
from ._PropertyExactlyParameterizer import PropertyExactlyParameterizer

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "PropertyAtLeastParameterizer",
    "PropertyAtMostParameterizer",
    "PropertyExactlyParameterizer",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import itertools as it

import pytest

from hstrat._auxiliary_lib import all_same
from hstrat.hstrat import HereditaryStratigraphicColumn, pseudostochastic_algo


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_impl_consistency(hash_salt):
    policy = pseudostochastic_algo.Policy(hash_salt)
    spec = policy.GetSpec()
    impls = [*pseudostochastic_algo._GenDropRanks_impls]
    instances = [impl(spec) for impl in impls]
    column = HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
    )
    for num_strata_deposited in range(1, 10**3):
        assert all_same(
            it.chain(
                (
                    sorted(
                        impl(spec)(
                            policy,
                            num_strata_deposited,
                            column.IterRetainedRanks(),
                        )
                    )
                    for impl in impls
                ),
                (
                    sorted(
                        instance(
                            policy,
                            num_strata_deposited,
                            column.IterRetainedRanks(),
                        )
                    )
                    for instance in instances
                ),
            )
        )
        column.DepositStratum()


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_eq(hash_salt):
    policy = pseudostochastic_algo.Policy(hash_salt)
    spec = policy.GetSpec()
    instance = pseudostochastic_algo.GenDropRanks(spec)

    assert instance == instance
    assert instance == pseudostochastic_algo.GenDropRanks(spec)
    assert instance is not None



"""Visualizations for stratum retention policies."""

from . import animate, ascii, plot
from .animate import *  # noqa: F401
from .ascii import *  # noqa: F401
from .plot import *  # noqa: F401

__all__ = animate.__all__ + ascii.__all__ + plot.__all__



import alifedata_phyloinformatics_convert as apc
import dendropy as dp
import pandas as pd


def load_dendropy_tree(path: str) -> dp.Tree:
    if path.endswith(".csv"):
        alifestd_df = pd.read_csv(path)
        return apc.alife_dataframe_to_dendropy_tree(
            alifestd_df,
            setup_edge_lengths=True,
        )
    elif path.endswith(".newick"):
        return dp.Tree.get(path=path, schema="newick")



from ._calc_common_ratio import calc_common_ratio
from ._calc_rank_backstop import calc_rank_backstop
from ._calc_rank_cutoff import calc_rank_cutoff
from ._calc_rank_sep import calc_rank_sep
from ._calc_target_rank import calc_target_rank
from ._calc_target_recency import calc_target_recency
from ._get_retained_ranks import get_retained_ranks

__all__ = [
    "calc_common_ratio",
    "calc_rank_backstop",
    "calc_rank_cutoff",
    "calc_rank_sep",
    "calc_target_rank",
    "calc_target_recency",
    "get_retained_ranks",
]



import collections
import itertools as it
import typing


# adapted from https://docs.python.org/3/library/itertools.html
def consume(iterator: typing.Iterator, n: int = None) -> None:
    "Advance the iterator n-steps ahead. If n is None, consume entirely."

    if not isinstance(iterator, collections.abc.Iterator):
        # consume is only effectual on an iterator, not an iterable
        # enforce against misconceived usage
        raise TypeError

    # Use functions that consume iterators at C speed.
    if n is None:
        # feed the entire iterator into a zero-length deque
        collections.deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(it.islice(iterator, n, n), None)



from hstrat._auxiliary_lib import iter_slices


def test_iter_slices_len_seq_0():
    result = list(iter_slices(0, 10))
    assert result == []


def test_iter_slices_len_seq_equals_chunk_size():
    result = list(iter_slices(10, 10))
    assert result == [slice(0, 10)]


def test_iter_slices_len_seq_less_than_chunk_size():
    result = list(iter_slices(5, 10))
    assert result == [slice(0, 10)]


def test_iter_slices_len_seq_greater_than_chunk_size():
    result = list(iter_slices(15, 5))
    assert result == [slice(0, 5), slice(5, 10), slice(10, 15)]


def test_iter_slices_start_greater_than_0():
    result = list(iter_slices(15, 5, 5))
    assert result == [slice(5, 10), slice(10, 15)]


def test_iter_slices_chunk_size_larger_than_input_list():
    result = list(iter_slices(10, 20))
    assert result == [slice(0, 20)]


def test_iter_slices_chunk_size_equal_to_one():
    result = list(iter_slices(10, 1))
    assert result == [
        slice(0, 1),
        slice(1, 2),
        slice(2, 3),
        slice(3, 4),
        slice(4, 5),
        slice(5, 6),
        slice(6, 7),
        slice(7, 8),
        slice(8, 9),
        slice(9, 10),
    ]


def test_iter_slices_start_parameter():
    result = list(iter_slices(12, 5, 2))
    assert result == [slice(2, 7), slice(7, 12)]


def test_iter_slices_chunk_size_does_not_divide_input_length():
    result = list(iter_slices(11, 5))
    assert result == [slice(0, 5), slice(5, 10), slice(10, 15)]


def test_iter_slices_chunk_size_equal_to_input_length():
    result = list(iter_slices(10, 10))
    assert result == [slice(0, 10)]


def test_iter_slices_chunk_size_does_not_divide_input_length_with_start():
    result = list(iter_slices(11, 5, 1))
    assert result == [slice(1, 6), slice(6, 11)]


def test_iter_slices_chunk_size_equal_to_input_length_with_start():
    result = list(iter_slices(10, 10, 1))
    assert result == [slice(1, 11)]


def test_iter_slices_chunk_size_divides_input_length_with_start_2():
    result = list(iter_slices(10, 5, 2))
    assert result == [slice(2, 7), slice(7, 12)]


def test_iter_slices_start_equals_length():
    result = list(iter_slices(10, 5, 10))
    assert result == []



import itertools as it
import os
import random

from Bio.Phylo.TreeConstruction import BaseTree
import alifedata_phyloinformatics_convert as apc
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import alifestd_validate

assets_path = os.path.join(os.path.dirname(__file__), "assets")

from . import _impl as impl


@pytest.mark.parametrize(
    "version_pin",
    [hstrat.__version__],
)
def test_empty_population(version_pin):
    population = []
    tree = hstrat.build_tree(
        population,
        version_pin=version_pin,
    )

    assert len(tree) == 0
    assert alifestd_validate(tree)


@pytest.mark.parametrize(
    "version_pin",
    [hstrat.__version__],
)
def test_dual_population_no_mrca(version_pin):
    organism1 = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(100)
    organism2 = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(100)

    population = [organism1, organism2]
    names = ["foo", "bar"]

    with pytest.raises(ValueError):
        tree = hstrat.build_tree(population, version_pin, taxon_labels=names)

    tree = hstrat.build_tree(
        population, version_pin, taxon_labels=names, force_common_ancestry=True
    )
    assert alifestd_validate(tree)
    tree.loc[:, "name"] = tree["taxon_label"]

    root_clade = BaseTree.Clade(name="Inner1")
    root_clade.clades = [
        BaseTree.Clade(branch_length=101.0, name="bar"),
        BaseTree.Clade(branch_length=101.0, name="foo"),
    ]
    true_tree = BaseTree.Tree(rooted=False, root=root_clade)

    assert (
        impl.tree_unweighted_robinson_foulds_distance(
            apc.alife_dataframe_to_biopython_tree(tree), true_tree
        )
        == 0.0
    )


@pytest.mark.parametrize(
    "version_pin",
    [hstrat.__version__],
)
def test_dual_population_with_mrca(version_pin):
    organism1 = hstrat.HereditaryStratigraphicColumn()
    organism2 = hstrat.HereditaryStratigraphicColumn()

    population = [organism1, organism2]
    names = ["foo", "bar"]

    for _ in range(100):
        parents = random.choices(population, k=len(population))
        population = [parent.CloneDescendant() for parent in parents]

    tree = hstrat.build_tree(
        population, version_pin=version_pin, taxon_labels=names
    )
    assert alifestd_validate(tree)
    tree["name"] = tree["taxon_label"]

    root_clade = BaseTree.Clade(name="Inner")
    root_clade.clades = [
        BaseTree.Clade(branch_length=0.0, name="bar"),
        BaseTree.Clade(branch_length=0.0, name="foo"),
    ]
    true_tree = BaseTree.Tree(rooted=False, root=root_clade)

    assert (
        impl.tree_unweighted_robinson_foulds_distance(
            apc.alife_dataframe_to_biopython_tree(tree),
            true_tree,
        )
        == 0.0
    )


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(3),
    ],
)
def test_reconstructed_mrca(orig_tree, retention_policy):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    reconst_df = hstrat.build_tree(extant_population, hstrat.__version__)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    pdm = reconst_tree.phylogenetic_distance_matrix()

    sorted_leaf_nodes = sorted(
        reconst_tree.leaf_node_iter(), key=lambda x: int(x.taxon.label)
    )
    assert len(sorted_leaf_nodes) == len(extant_population)
    assert [
        leaf_node.distance_from_root() for leaf_node in sorted_leaf_nodes
    ] == [
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    ]

    for reconst_node_pair, extant_column_pair in zip(
        it.combinations(sorted_leaf_nodes, 2),
        it.combinations(extant_population, 2),
    ):
        reconst_mrca = impl.descend_unifurcations(
            pdm.mrca(*map(lambda x: x.taxon, reconst_node_pair))
        )

        (
            lower_mrca_bound,
            upper_mrca_bound,
        ) = hstrat.calc_rank_of_mrca_bounds_between(
            *extant_column_pair, prior="arbitrary"
        )

        assert (
            lower_mrca_bound
            <= reconst_mrca.distance_from_root()
            < upper_mrca_bound
        )


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
@pytest.mark.parametrize(
    "version_pin",
    [hstrat.__version__],
)
def test_determinism(orig_tree, retention_policy, wrap, version_pin):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    first_reconst = hstrat.build_tree(extant_population, version_pin)
    for _rep in range(3):
        _ = _rep
        second_reconst = hstrat.build_tree(
            [wrap(col) for col in extant_population], version_pin
        )
        assert first_reconst.equals(second_reconst)


@pytest.mark.parametrize(
    "version_pin",
    [hstrat.__version__],
)
@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(4),
    ],
)
def test_reconstructed_taxon_labels(version_pin, orig_tree, retention_policy):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )
    taxon_labels = [str(id(x)) for x in extant_population]

    reconst_df = hstrat.build_tree(
        extant_population,
        version_pin,
        taxon_labels=taxon_labels,
    )
    assert "taxon_label" in reconst_df
    assert set(taxon_labels) < set(reconst_df["taxon_label"])

    reconst_df = hstrat.build_tree(
        extant_population,
        version_pin,
    )
    assert "taxon_label" in reconst_df
    assert len(reconst_df["taxon_label"].unique()) == len(
        reconst_df["taxon_label"]
    )
    assert set(map(str, range(len(extant_population)))) < set(
        reconst_df["taxon_label"]
    )



import typing


def alifestd_parse_ancestor_id(ancestor_list_str: str) -> typing.Optional[int]:
    """Parse at most a single ancestor id from an `ancestor_list` field."""
    if ancestor_list_str.lower() in (
        "[]",
        "[none]",
    ):
        return None
    else:
        without_brackets_str = ancestor_list_str[1:-1]
        return int(without_brackets_str)



import numbers
import typing

import networkx as nx

from ._estimate_origin_times import estimate_origin_times


def find_chronological_roots(
    graph: nx.Graph,
    leaf_node_origin_times: typing.Dict[typing.Any, numbers.Number],
) -> typing.List[str]:
    """Work backward from known phylogenetic depths of leaf nodes to identify
    a likely most ancient ancestor.

    Keys in `leaf_node_origin_times` should correspond to nodes in graph.
    """

    node_origin_times = estimate_origin_times(graph, leaf_node_origin_times)

    # extract root node of each connected component
    root_nodes = [
        min(component, key=lambda node: node_origin_times[node])
        for component in nx.connected_components(graph)
    ]

    return [graph.nodes[root_node]["taxon_label"] for root_node in root_nodes]



import anytree
import pytest

import hstrat.phylogenetic_inference.tree._impl as impl


@pytest.fixture
def inner_node():
    return anytree.AnyNode()


def test_taxon_label(inner_node):
    leaf_node = impl.TrieLeafNode(inner_node, "taxon")
    assert leaf_node.taxon_label == "taxon"


def test_taxon(inner_node):
    leaf_node = impl.TrieLeafNode(inner_node, "taxon")
    assert leaf_node.taxon == "taxon"


def test_parent(inner_node):
    leaf_node = impl.TrieLeafNode(inner_node, "taxon")
    assert leaf_node.parent is inner_node
